#!/usr/bin/env python
import ursgal
import os
import csv
import sys


class novor_1_05( ursgal.UNode ):
    """
    Novor UNode
    Parameter options at http://rapidnovor.com/

    Reference:
    Bin Ma (2015) Novor: Real-Time Peptide de Novo Sequencing Software. J Am Soc Mass Spectrom 26 (11)

    Import node for version novor_1_1beta

    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Novor',
        'version'            : '1.05',
        'release_date'       : '2015-11-10',
        'engine_type' : {
            'de_novo_search_engine' : True,
        },
        'input_extensions'   : ['.mgf'],
        'output_extensions'  : ['.csv'],
        'in_development'     : False,
        'create_own_folder'  : True,
        'include_in_git'     : False,
        'distributable'      : False,
        'utranslation_style' : 'novor_style_1',
        'engine' : {
            'linux' : {
                '64bit' : {
                    'exe' : 'novor.sh',
                }
            },
            'darwin' : {
                '64bit' : {
                    'exe' : 'novor.sh',
                }
            },
            'win32' : {
                '64bit' : {
                    'exe' : 'novor.bat',
                }
            },
        },
        'citation' : \
            'Bin Ma (2015) Novor: Real-Time Peptide de Novo Sequencing '\
            'Software. J Am Soc Mass Spectrom 26 (11)',
    }

    def __init__(self, *args, **kwargs):
        super(novor_1_05, self).__init__(*args, **kwargs)
        pass


    def preflight( self ):
        '''
        Formatting the command line via self.params

        Params.txt file will be created in the output folder

        Returns:
                dict: self.params
        '''

        self.params['translations']['mgf_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        self.params['translations']['mgf_new_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['file_root'] + '_tmp.mgf'
        )
        self.created_tmp_files.append( self.params['translations']['mgf_new_input_file'] )

        self.params['translations']['tmp_output_file_incl_path'] = os.path.join(
            self.params['translations']['mgf_new_input_file'] + '.csv'
        )
        self.created_tmp_files.append( self.params['translations']['tmp_output_file_incl_path'] )

        self.params['translations']['params_file'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file'] + '_Params.txt'
        )
        self.created_tmp_files.append( self.params['translations']['params_file'])

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        print(
            '''
            [ WARNING ] precursor_mass_tolerance_plus and precursor_mass_tolerance_minus
            [ WARNING ] need to be combined for Novor (use of symmetric tolerance window).
            [ WARNING ] The arithmetic mean is used.
            '''
        )
        self.params['translations']['precursor_mass_tolerance'] = ( float(self.params['precursor_mass_tolerance_plus']) + \
                                                            float(self.params['precursor_mass_tolerance_minus']) ) \
                                                            / 2.0

        available_mods =[
            'Acetyl (K)',
            'Acetyl (N-term)',
            'Amidated (C-term)',
            # 'Ammonia-loss (N-term C)',
            'Biotin (K)',
            'Biotin (N-term)',
            'Carbamidomethyl (C)',
            'Carbamyl (K)',
            'Carbamyl (N-term)',
            'Carboxymethyl (C)',
            # 'Deamidated (NQ)',
            # 'Dehydrated (N-term C)',
            'Dioxidation (M)',
            'Methyl (C-term)',
            'Methyl (DE)',
            'Oxidation (M)',
            'Oxidation (HW)',
            'Phospho (ST)',
            'Phospho (Y)',
            # 'Pyro-carbamidomethyl (N-term C)',
            'Pyro-Glu (E)',
            'Pyro-Glu (Q)',
            'Sodium (C-term)',
            'Sodium (DE)',
            'Sulfo (STY)',
            'Trimethyl (RK)',
        ]

        # need to check with mass
        potential_mods = {'tag': 'opt', 'mods' : []}
        fixed_mods = {'tag': 'fix', 'mods' : []}
        for mod_type in [potential_mods, fixed_mods]:
            not_available_mods = {}
            for mod in self.params[ 'mods' ][ mod_type['tag'] ]:
                if 'N-term' in mod['pos']:
                    mod[ 'aa' ] = 'N-term'
                if 'C-term' in mod['pos']:
                    mod[ 'aa' ] = 'C-term'
                if '{0} ({1})'.format(mod['name'], mod['aa']) not in available_mods:
                    if mod['name'] not in not_available_mods.keys():
                        not_available_mods[mod['name']] = []
                    not_available_mods[mod['name']].append(mod['aa'])
                    continue
                mod_type['mods'].append('{0} ({1})'.format(mod[ 'name' ], mod[ 'aa' ] ))

            for mod in not_available_mods.keys():
                if '{0} ({1})'.format(mod, ''.join(sorted(not_available_mods[mod]))) not in available_mods:
                    print('''
            [ WARNING ] Novor does not support your given modification
            [ WARNING ] Continue without modification {0} ({1})'''.format(mod, ''.join(sorted(not_available_mods[mod])))
                        )
                    continue
                else:
                    mod_type['mods'].append('{0} ({1})'.format(mod, ''.join(sorted(not_available_mods[mod]))))

        self.params['translations']['fixed_modifications'] =  ','.join( fixed_mods['mods'] )
        self.params['translations']['potential_modifications'] = ','.join( potential_mods['mods'] )

        if self.params['translations']['frag_mass_tolerance_unit'] == 'ppm':
            self.params['translations']['frag_mass_tolerance'] = \
                ursgal.ucore.convert_ppm_to_dalton(
                    self.params['translations']['frag_mass_tolerance'],
                    base_mz=self.params['translations']['base_mz']
                )

        params_file = open( self.params['translations']['params_file'], 'w', encoding = 'UTF-8' )
        params2write = [
            'enzyme = {enzyme}'.format(**self.params['translations']),
            'fragmentation = {frag_method}'.format(**self.params['translations']),
            'massAnalyzer = {instrument}'.format(**self.params['translations']),
            'fragmentIonErrorTol = {frag_mass_tolerance}Da'.format(**self.params['translations']),
            'precursorErrorTol = {precursor_mass_tolerance}{precursor_mass_tolerance_unit}'.format(**self.params['translations']),
            'variableModifications = {potential_modifications}'.format(**self.params['translations']),
            'fixedModifications = {fixed_modifications}'.format(**self.params['translations']),
            'forbiddenResidues = {forbidden_residues}'.format(**self.params['translations']),
        ]
        for param in params2write:
            print( param, file = params_file )
        params_file.close()

        mgf_org_input_file = open( self.params['translations']['mgf_input_file'], 'r', encoding = 'UTF-8' )
        lines = mgf_org_input_file.readlines()
        mgf_org_input_file.close()

        mgf_new_input_file = open( self.params['translations']['mgf_new_input_file'], 'w', encoding = 'UTF-8')
        for line in lines:
            if line.startswith('CHARGE'):
                charge = line.split('=')[1]
                print( 'CHARGE={0}+'.format(charge.strip()), file = mgf_new_input_file)
            else:
                print( line, file = mgf_new_input_file)
        mgf_new_input_file.close()

        self.params[ 'command_list' ] = [
            self.exe,
            '-p', '{params_file}'.format(**self.params['translations']),
            self.params['translations']['mgf_new_input_file'], '-f',
        ]

        return self.params

    def postflight( self ):
        '''
        Reformats the Novor output file
        '''
        regex_list = [
            ('\(Cam\)','Carbamidomethyl'),
            ('\(O\)','Oxidation'),
            ('\(N-term Acetyl\)','Acetyl'),
            ('\[Acetyl\]','Acetyl'),
            ('\(Acetyl\)','Acetyl'),
            ('\[Amidated\]','Amidated'),
            ('\[Biotin\]','Biotin'),
            ('\(Biotin\)','Biotin'),
            ('\[Carbamyl\]','Carbamyl'),
            ('\(Carbamyl\)','Carbamyl'),
            ('\(Carboxymethyl\)','Carboxymethyl'),
            ('\(Phospho\)','Phospho'),
            ('\(Deamidated\)', 'Deamidated'),
            ('\(Dioxidation\)', 'Dioxidation'),
            ('\[Methyl\]','Methyl'),
            ('\(Methyl\)','Methyl'),
            ('\(Pyro-Glu\)','Pyro-Glu'),
            ('\[Sodium\]','Sodium'),
            ('\(Sodium\)','Sodium'),
            ('\(Sulfo\)','Sulfo'),
            ('\(Trimethyl\)','Trimethyl'),
        ]

        cached_novor_output = []
        headers = None
        result_file = open( self.params['translations']['tmp_output_file_incl_path'], 'r')
        for line in result_file:
            if line.startswith('#'):
                if line.startswith('# id'):
                    headers = line.strip().split(',')
                    break
        assert headers is not None , '''Change in NOVOR output ? could not find header starting wth # id'''

        print('[ PARSING  ] Loading unformatted Novor results ...')
        non_commented_csv_lines = [ row for row in result_file if not row.startswith('#') ]
        result_file.close()
        csv_dict_reader_object = csv.DictReader(
            [ ','.join(headers) ] + non_commented_csv_lines
        )
        # extend and tanslate headers
        translated_headers = []
        header_translations = self.UNODE_UPARAMS['header_translations']['uvalue_style_translation']
        for header in headers:
            if header == '':
                continue
            translated_headers.append(
                    header_translations.get(header, header)
                )
        translated_headers += [
            'Spectrum Title',
            'Modifications',
            'Raw data location',
        ]

        new_line_dict_list = []
        for line_dict in csv_dict_reader_object:
            new_line_dict = {}
            for k, v in line_dict.items():
                if v is None:
                    continue
                if header_translations[k] == 'Sequence':
                    formated_peptide = v
                    for regex_pattern, unimod_name in regex_list:
                        formated_peptide = ursgal.ucore.reformat_peptide(
                            regex_pattern,
                            unimod_name,
                            formated_peptide
                        )
                    new_line_dict['Sequence'] = formated_peptide.split('#')[0]
                    try:
                        new_line_dict['Modifications'] = formated_peptide.split('#')[1]
                    except:
                        new_line_dict['Modifications'] = ''
                    continue
                new_line_dict[ header_translations[k] ] = v.strip()
            new_line_dict['Raw data location'] = self.params['translations']['mgf_input_file']
            new_line_dict_list.append(new_line_dict)

        new_result_file = open( self.params['translations']['output_file_incl_path'], 'w')
        csv_kwargs = {}
        if sys.platform == 'win32':
            csv_kwargs['lineterminator'] = '\n'
        else:
            csv_kwargs['lineterminator'] = '\r\n'
        csv_dict_writer_object = csv.DictWriter(
            new_result_file,
            fieldnames = translated_headers,
            **csv_kwargs
        )
        csv_dict_writer_object.writeheader()
        for line_dict in new_line_dict_list:
            csv_dict_writer_object.writerow( line_dict )
        return


#!/usr/bin/env python3
# encoding: utf-8
import ursgal
import os
import sys
import pprint


def main(folder):
    '''

    usage:

        ./human_br_complete_workflow.py <folder_with_human_br_files>

    This scripts produces the data for figure 3.

    '''

    # Initialize the UController:
    uc = ursgal.UController(
        params={
            'enzyme': 'trypsin',
            'decoy_generation_mode': 'reverse_protein',
        }
    )

    # MS Spectra, downloaded from http://proteomecentral.proteomexchange.org
    # via the dataset accession PXD000263 and converted to mzML

    mass_spec_files = [
        '120813OTc1_NQL-AU-0314-LFQ-LCM-SG-01_013.mzML',
        '120813OTc1_NQL-AU-0314-LFQ-LCM-SG-02_025.mzML',
        '120813OTc1_NQL-AU-0314-LFQ-LCM-SG-03_033.mzML',
        '120813OTc1_NQL-AU-0314-LFQ-LCM-SG-04_048.mzML',
    ]

    for mass_spec_file in mass_spec_files:
        if os.path.exists(os.path.join(folder, mass_spec_file)) is False:
            print(
                'Please download RAW files to folder {} and convert to mzML:'.format(folder))
            pprint.pprint(mass_spec_files)
            sys.exit(1)

    # mods from Wen et al. (2015):
    modifications = [
        # Carbamidomethyl  (C) was set as fixed modification
        'C,fix,any,Carbamidomethyl',
        'M,opt,any,Oxidation',        # Oxidation (M) as well as
        # Deamidated (NQ) were set as optional modification
        'N,opt,any,Deamidated',
        # Deamidated (NQ) were set as optional modification
        'Q,opt,any,Deamidated',
    ]

    # The target peptide database which will be searched (UniProt Human
    # reference proteome from July 2013)
    target_database = 'uniprot_human_UP000005640_created_until_20130707.fasta'
    # Let's turn it into a target decoy database by reversing peptides:
    target_decoy_database = uc.execute_misc_engine(
        input_file=target_database,
        engine='generate_target_decoy_1_0_0'
    )

    # OMSSA parameters from Wen et al. (2015):
    omssa_params = {
        # (used by default) # -w
        'he': '1000', # -he 1000
        'zcc': '1', # -zcc 1
        'frag_mass_tolerance': '0.6', # -to 0.6
        'frag_mass_tolerance_unit': 'da', # -to 0.6
        'precursor_mass_tolerance_minus': '10', # -te 10
        'precursor_mass_tolerance_plus': '10', # -te 10
        'precursor_mass_tolerance_unit': 'ppm', # -teppm
        'score_a_ions': False, # -i 1,4
        'score_b_ions': True, # -i 1,4
        'score_c_ions': False, # -i 1,4
        'score_x_ions': False, # -i 1,4
        'score_y_ions': True, # -i 1,4
        'score_z_ions': False, # -i 1,4
        'enzyme': 'trypsin_p', # -e 10
        'maximum_missed_cleavages': '1', # -v 1
        'precursor_max_charge': '8', # -zh 8
        'precursor_min_charge': '1', # -zl 1
        'tez': '1', # -tez 1
        'precursor_isotope_range': '0,1', # -ti 1
        'num_match_spec': '1', # -hc 1

        'database': target_decoy_database,
        'modifications': modifications,
    }

    # MS-GF+ parameters from Wen et al. (2015):
    msgf_params = {
        # precursor ion mass tolerance was set to 10 ppm
        'precursor_mass_tolerance_unit': 'ppm',
        # precursor ion mass tolerance was set to 10 ppm
        'precursor_mass_tolerance_minus': '10',
        # precursor ion mass tolerance was set to 10 ppm
        'precursor_mass_tolerance_plus': '10',
        # the max number of optional modifications per peptide were set as 3
        # (used by default) # number of allowed isotope errors was set as 1
        'enzyme': 'trypsin', # the enzyme was set as trypsin
        # (used by default) # fully enzymatic peptides were specified, i.e. no non-enzymatic termini
        'frag_method': '1',  # the fragmentation method selected in the search was CID
        'max_pep_length': '45',  # the maximum peptide length to consider was set as 45
        # the minimum precursor charge to consider if charges are not specified
        # in the spectrum file was set as 1
        'precursor_min_charge': '1',
        # the maximum precursor charge to consider was set as 8
        'precursor_max_charge': '8',
        # (used by default) # the parameter 'addFeatures' was set as 1 (required for Percolator)
        # all of the other parameters were set as default
        # the instrument selected
        # was High-res

        'database': target_decoy_database,
        'modifications': modifications,
    }

    # X!Tandem parameters from Wen et al. (2015):
    xtandem_params = {
        # precursor ion mass tolerance was set to 10 ppm
        'precursor_mass_tolerance_unit': 'ppm',
        # precursor ion mass tolerance was set to 10 ppm
        'precursor_mass_tolerance_minus': '10',
        # precursor ion mass tolerance was set to 10 ppm
        'precursor_mass_tolerance_plus': '10',
        # the fragment ion mass tolerance was set to 0.6 Da
        'frag_mass_tolerance': '0.6',
        # the fragment ion mass tolerance was set to 0.6 Da
        'frag_mass_tolerance_unit': 'da',
        # parent monoisotopic mass isotope error was set as 'yes'
        'precursor_isotope_range': '0,1',
        'precursor_max_charge': '8', # maximum parent charge of spectrum was set as 8
        'enzyme': 'trypsin',  # the enzyme was set as trypsin ([RK]|[X])
        # the maximum missed cleavage sites were set as 1
        'maximum_missed_cleavages': '1',
        # (used by default) # no model refinement was employed.

        'database': target_decoy_database,
        'modifications': modifications,
    }

    search_engine_settings = [
        # not used in Wen et al., so we use the same settings as xtandem
        ('msamanda_1_0_0_5243', xtandem_params, 'LTQ XL high res'),
        # not used in Wen et al., so we use the same settings as xtandem
        ('myrimatch_2_1_138', xtandem_params, 'LTQ XL high res'),
        # the instrument selected was High-res
        ('msgfplus_v9979', msgf_params, 'LTQ XL high res'),
        ('xtandem_jackhammer', xtandem_params, None),
        ('omssa_2_1_9', omssa_params, None),
    ]

    merged_validated_files_3_engines = []
    merged_validated_files_5_engines = []

    for engine, wen_params, instrument in search_engine_settings:

        # Initializing the uPLANIT UController class with
        # our specified modifications and mass spectrometer
        uc = ursgal.UController(
            params=wen_params
        )

        if instrument is not None:
            uc.set_profile(instrument)

        unified_results = []
        percolator_validated_results = []

        for mzML_file in mass_spec_files:
            unified_search_results = uc.search(
                input_file=mzML_file,
                engine=engine,
            )
            unified_results.append(
                unified_search_results
            )
            validated_csv = uc.validate(
                input_file=unified_search_results,
                engine='percolator_2_08',
            )
            percolator_validated_results.append(validated_csv)

        merged_validated_csv = uc.execute_misc_engine(
            input_file=percolator_validated_results,
            engine='merge_csvs_1_0_0'
        )
        merged_unvalidated_csv = uc.execute_misc_engine(
            input_file=unified_results,
            engine='merge_csvs_1_0_0',
        )

        if engine in ["omssa_2_1_9", "xtandem_jackhammer", "msgfplus_v9979"]:
            merged_validated_files_3_engines.append(merged_validated_csv)
        merged_validated_files_5_engines.append(merged_validated_csv)

    uc.params['prefix'] = '5-engines-summary'
    uc.combine_search_results(
        input_files=merged_validated_files_5_engines,
        engine='combine_FDR_0_1',
    )

    uc.params['prefix'] = '3-engines-summary'
    uc.combine_search_results(
        input_files=merged_validated_files_3_engines,
        engine='combine_FDR_0_1',
    )


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(main.__doc__)
        sys.exit(1)
    main(sys.argv[1])



import csv
import os
import pickle
import ursgal


test_case_list = [
    ('TMT_xtandem_vengeance_pmap.csv', 'xtandem_vengeance'),
    ('TMT_msgfplus_pmap.csv', 'msgfplus_v2018_09_12')
]

params = {
    'modifications': [
        'M,opt,any,Oxidation',
        '*,opt,Prot-N-term,Acetyl',
        'C,fix,any,Carbamidomethyl',
        '*,opt,N-term,TMT6plex',
        'K,fix,any,TMT6plex',
    ],
}
# uc = ursgal.UController(params=params, profile='QExactive+', verbose=False)
# uc.scan_rt_lookup_path = 'tests/data/_ursgal_test_lookup.pkl'
# uc.map_mods()
# unify_csv_main = uc.unodes['unify_csv_1_0_0']['class'].import_engine_as_python_function()


def unify_nterm_test():
    for param_l in test_case_list:
        yield unify_csv, param_l[0], param_l[1]


def unify_csv(file, engine):
    uc = ursgal.UController(params=params, profile='QExactive+', verbose=False)
    uc.scan_rt_lookup_path = 'tests/data/_ursgal_test_lookup.pkl'
    uc.map_mods()
    unify_csv_main = uc.unodes['unify_csv_1_0_0']['class'].import_engine_as_python_function()

    output_csv = os.path.join(
        'tests', 'data',
        os.path.splitext(file)[0] + '_unified.csv'
    )
    input_csv = os.path.join(
        'tests', 'data',
        file
    )
    scan_rt_lookup = pickle.load(
        open(
            os.path.join(
                'tests',
                'data',
                '_ursgal_test_pickle.pkl')
            ,
            'rb'
        )
    )
    unify_csv_main(
        input_file=input_csv,
        output_file=output_csv,
        scan_rt_lookup=scan_rt_lookup,
        params={
            'translations': {
                'decoy_tag': 'decoy_',
                'enzyme': 'KR;C;P',
                'semi_enzyme': False,
                'database': os.path.join('tests', 'data', 'P0ADZ4.fasta'),
                'protein_delimiter': '<|>',
                'psm_merge_delimiter': ';',
                'keep_asp_pro_broken_peps': True,
                'precursor_mass_tolerance_minus': 5,
                'precursor_mass_tolerance_plus': 5,
                'precursor_isotope_range': "0,1",
                'max_missed_cleavages': 2,
                'rounded_mass_decimals': 3,
                'use_pyqms_for_mz_calculation': False,
                'aa_exception_dict': {
                    'J': {
                        'original_aa': ['L', 'I'],
                    },
                    'O': {
                        'original_aa': ['K'],
                        'unimod_name': 'Methylpyrroline',
                    },
                },
            },
            'label': '',
            'mods': uc.params['mods'],
        },
        search_engine=engine,
    )
    reader_produced = [line for line in csv.DictReader(open(output_csv))]
    reader_expected = [line for line in csv.DictReader(open(output_csv+'_expected.csv'))]
    for pos, line in enumerate(reader_produced):
        print('#{pos:0>5} Produced: {mod}'.format(pos=pos, mod=line["Modifications"]))
        print('#{pos:0>5} Expected: {mod}'.format(pos=pos, mod=reader_expected[pos]["Modifications"]))
        assert line['Modifications'] == reader_expected[pos]['Modifications']


if __name__ == '__main__':
    for i, params_l in enumerate(test_case_list):
        unify_csv(params_l[0], params_l[1])



#!/usr/bin/env python
import ursgal
import os
import csv
import sys


class deepnovo_0_0_1(ursgal.UNode):
    """
    DeepNovo UNode
    For further information, see https://github.com/nh2tran/DeepNovo

    Note:
        Please download manually from https://github.com/StSchulze/DeepNovo?organization=StSchulze&organization=StSchulze
        or using git clone https://github.com/StSchulze/DeepNovo.git
        and download the model from https://drive.google.com/open?id=0By9IxqHK5MdWalJLSGliWW1RY2c

    Reference:
    Tran, N.H.; Zhang, X.; Xin, L.; Shan, B.; Li, M. (2017) De novo peptide sequencing by deep learning. PNAS 114 (31)

    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'DeepNovo',
        'version'            : '0.0.1',
        'release_date'       : '2017-11-29',
        'engine_type' : {
            'de_novo_search_engine' : True,
        },
        'input_extensions'   : ['.mgf'],
        'output_extensions'  : ['.csv'],
        'in_development'     : False,
        'create_own_folder'  : True,
        'include_in_git'     : False,
        'distributable'      : False,
        'utranslation_style' : 'deepnovo_style_1',
        'engine' : {
            'platform_independent'    : {
                'arc_independent' : {
                    'exe'            : 'deepnovo_main.py',
                    'url'            : 'https://github.com/nh2tran/DeepNovo',
                    'zip_md5'        : '',
                    'additional_exe' : [],
                },
            },
        },
        'citation' :
        'Tran, N.H.; Zhang, X.; Xin, L.; Shan, B.; Li, M. (2017) '
            'De novo peptide sequencing by deep learning. PNAS 114 (31)'
    }

    def __init__(self, *args, **kwargs):
        super(deepnovo_0_0_1, self).__init__(*args, **kwargs)
        pass

    def preflight(self):
        """
        Create deepnovo_config.py file and format command line via self.params

        Returns:
                dict: self.params
        """

        self.time_point(tag='execution')

        # main = self.import_engine_as_python_function()

        self.params['translations']['mgf_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        self.params['translations']['mgf_new_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['file_root'] + '_tmp.mgf'
        )
        self.created_tmp_files.append(
            self.params['translations']['mgf_new_input_file']
        )

        self.params['translations']['tmp_output_file_incl_path'] = os.path.join(
            self.params['translations']['mgf_new_input_file'] + '.tsv'
        )
        # self.created_tmp_files.append(
        #     self.params['translations']['tmp_output_file_incl_path']
        # )

        self.params['translations']['params_file'] = os.path.join(
            os.path.dirname(self.exe),
            'deepnovo_config.py'
        )
        # self.created_tmp_files.append(
        #     self.params['translations']['params_file'])

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        mgf_org_input_file = open(
            self.params['translations']['mgf_input_file'], 'r', encoding='UTF-8'
        )
        lines = mgf_org_input_file.readlines()
        mgf_org_input_file.close()

        self.scan_lookup = {}
        print('rewriting mgf input file to include SEQ')
        with open(
            self.params['translations']['mgf_new_input_file'], 'w', encoding='UTF-8'
        ) as mgf_new_input_file:
            for n, line in enumerate(lines):
                line = line.strip()
                if line.startswith('BEGIN IONS'):
                    entry = [line]
                    entry_dict = {}
                elif line.startswith('TITLE='):
                    entry_dict['TITLE'] = line
                elif line.startswith('SEQ='):
                    entry_dict['SEQ'] = line
                elif line.startswith('PEPMASS='):
                    entry_dict['PEPMASS'] = line
                elif line.startswith('CHARGE='):
                    entry_dict['CHARGE'] = line
                elif line.startswith('SCANS='):
                    entry_dict['SCANS'] = line
                elif line.startswith('RTINSECONDS='):
                    entry_dict['RTINSECONDS'] = line
                elif line.startswith('END IONS'):
                    entry.append(line)
                    entry.append('')
                    # if 'SEQ' not in entry_dict:
                    #     entry_dict['SEQ'] = 'SEQ= '
                    scan = entry_dict['SCANS'].split('=')[1]
                    charge = entry_dict['CHARGE'].split('=')[1].strip('+')
                    self.scan_lookup[scan] = charge
                    for n, write_line in enumerate(entry):
                        if n == 1:
                            for header in [
                                'TITLE',
                                'PEPMASS',
                                'CHARGE',
                                'SCANS',
                                'RTINSECONDS',
                                # 'SEQ',
                            ]:
                                print(
                                    entry_dict[header],
                                    file=mgf_new_input_file
                                )
                        print(write_line, file=mgf_new_input_file)
                else:
                    entry.append(line)
        mgf_new_input_file.close()

        print(
            '''
            [ WARNING ] precursor_mass_tolerance_plus and precursor_mass_tolerance_minus
            [ WARNING ] need to be combined for DeepNovo (use of symmetric tolerance window).
            [ WARNING ] The arithmetic mean is used.
            '''
        )

        self.params_to_write = {
            'mgf_new_input_file' : self.params['translations']['mgf_new_input_file'],
            'output_path' : self.params['translations']['tmp_output_file_incl_path'],
        }
        self.params['translations']['precursor_mass_tolerance'] = (float(self.params['precursor_mass_tolerance_plus']) +
                                                                   float(self.params['precursor_mass_tolerance_minus']) ) \
            / 2.0

        if self.params['translations']['precursor_mass_tolerance_unit'] == 'ppm':
            self.params_to_write['precursor_mass_tolerance_da'] = \
                ursgal.ucore.convert_ppm_to_dalton(
                    self.params['translations']['precursor_mass_tolerance'],
                    base_mz=self.params['translations']['base_mz']
            )
            self.params_to_write['precursor_mass_tolerance_ppm'] = \
                self.params['translations']['precursor_mass_tolerance']
        elif self.params['translations']['precursor_mass_tolerance_unit'] == 'da':
            self.params_to_write['precursor_mass_tolerance_ppm'] = \
                ursgal.ucore.convert_dalton_to_ppm(
                    self.params['translations']['precursor_mass_tolerance'],
                    base_mz=self.params['translations']['base_mz']
            )
            self.params_to_write['precursor_mass_tolerance_da'] = \
                self.params['translations']['precursor_mass_tolerance']

        assert self.params['translations']['deepnovo_mode'] == 'search_denovo', '''
            [ ERROR ] Only search_denovo supported as deepnovo_mmode so far!
            '''
        # self.params_to_write['denovo_mode'] = False
        # self.params_to_write['db_mode'] = False
        # self.params_to_write['hybrid_mode'] = False
        # if self.params['translations']['deepnovo_mode'] == 'search_denovo':
        #     self.params_to_write['denovo_mode'] = True
        # elif self.params['translations']['deepnovo_mode'] == 'search_db':
        #     self.params_to_write['db_mode'] = True
        # elif self.params['translations']['deepnovo_mode'] == 'search_hybrid':
        #     self.params_to_write['hybrid_mode'] = True

        # vocab_reverse = [
        #     'A',
        #     'R',
        #     'N',
        #     'D',
        #     'C',
        #     'E',
        #     'Q',
        #     'G',
        #     'H',
        #     'I',
        #     'L',
        #     'K',
        #     'M',
        #     'F',
        #     'P',
        #     'S',
        #     'T',
        #     'W',
        #     'Y',
        #     'V',
        # ]
        # mass_H = 1.0078
        # mass_N_terminus = 1.0078
        # mass_C_terminus = 17.0027
        # mass_AA = dict([
        #     ('_PAD', 0.0),
        #     ('_GO', mass_N_terminus-mass_H),
        #     ('_EOS', mass_C_terminus+mass_H),
        #     ('A', 71.03711),
        #     ('R', 156.10111),
        #     ('N', 114.04293),
        #     ('D', 115.02694),
        #     ('C', 103.00919),
        #     ('E', 129.04259),
        #     ('Q', 128.05858),
        #     ('G', 57.02146),
        #     ('H', 137.05891),
        #     ('I', 113.08406),
        #     ('L', 113.08406),
        #     ('K', 128.09496),
        #     ('M', 131.04049),
        #     ('F', 147.06841),
        #     ('P', 97.05276),
        #     ('S', 87.03203),
        #     ('T', 101.04768),
        #     ('W', 186.07931),
        #     ('Y', 163.06333),
        #     ('V', 99.06841),
        # ])
        # import pprint
        # pprint.pprint(self.params['translations'])
        # exit()
        for deepnovo_param in self.params['translations']['_grouped_by_translated_key'].keys():
            for ursgal_param_name, param_value in self.params['translations']['_grouped_by_translated_key'][deepnovo_param].items():
                if type(deepnovo_param) is tuple:
                    continue
                elif deepnovo_param == 'knapsack_file':
                    if param_value is None or param_value == 'default':
                        knapsack_file = os.path.join(
                            os.path.dirname(self.exe),
                            'knapsack.npy'
                        )
                    else:
                        knapsack_file = param_value
                    self.params_to_write['knapsack_file'] = knapsack_file
                elif deepnovo_param == 'train_dir':
                    if param_value is None or param_value == 'default':
                        train_dir = os.path.join(
                            os.path.dirname(self.exe),
                            'train.example'
                        )
                    else:
                        train_dir = param_value
                    self.params_to_write['train_dir'] = train_dir
                elif deepnovo_param == 'modifications':
                    assert set(param_value) == set(
                        ['M,opt,any,Oxidation',
                         'C,fix,any,Carbamidomethyl',
                         'N,opt,any,Deamidated',
                         'Q,opt,any,Deamidated']), '''
                    [ ERROR ] The default model of DeepNovo only supports the following modification list:
                        ['M,opt,any,Oxidation',
                         'C,fix,any,Carbamidomethyl',
                         'N,opt,any,Deamidated',
                         'Q,opt,any,Deamidated']
                    [ ERROR ] You specified instead: {0}
                    '''.format(param_value)
                #     cc = ursgal.ChemicalComposition()
                #     for mod_dict in self.params['mods']['opt']:
                #         '''
                #         {'_id': 0,
                #           'aa': '*',
                #           'composition': {'C': 2, 'H': 2, 'O': 1},
                #           'id': '1',
                #           'mass': 42.010565,
                #           'name': 'Acetyl',
                #           'org': '*,opt,Prot-N-term,Acetyl',
                #           'pos': 'Prot-N-term',
                #           'unimod': True},
                #         '''
                #         if mod_dict['pos'] != 'any':
                #             print(
                #                 '''
                #   [WARNING] DeepNovo does not support positional arguments for modifications.
                #   [WARNING] The following modification will be used on any position:
                #   [WARNING] {0}
                #             '''.format(mod_dict['org'])
                #             )
                #         if mod_dict['aa'] == '*':
                #             print('''
                #             Not sure how to handle this modification in DeepNovo:
                #             {0}
                #             '''.format(mod_dict['org']))
                #             sys.exit(1)
                #         mod_index = vocab_reverse.index(mod_dict['aa'])
                #         vocab_reverse.insert(mod_index+1, mod_dict['aa']+'mod')
                #         cc.use('{0}#{1}:1'.format(
                #             mod_dict['aa'],
                #             mod_dict['name']
                #         ))
                #         mass_AA[mod_dict['aa']+'mod'] = cc._mass()

                #     for mod_dict in self.params['mods']['fix']:
                #         if mod_dict['pos'] != 'any':
                #             print(
                #                 '''
                #   [WARNING] DeepNovo does not support positional arguments for modifications.
                #   [WARNING] The following modification will be used on any position:
                #   [WARNING] {0}
                #             '''.format(mod_dict['org'])
                #             )
                #         if mod_dict['aa'] == '*':
                #             print('''
                #             Not sure how to handle this modification in DeepNovo:
                #             {0}
                #             '''.format(mod_dict['org']))
                #             sys.exit(1)
                #         mod_index = vocab_reverse.index(mod_dict['aa'])
                #         vocab_reverse[mod_index] = mod_dict['aa']+'mod'
                #         del mass_AA[mod_dict['aa']]
                #         cc.use('{0}#{1}:1'.format(
                #             mod_dict['aa'],
                #             mod_dict['name']
                #         ))
                #         mass_AA[mod_dict['aa']+'mod'] = cc._mass()
                #     self.params_to_write['mass_AA'] = mass_AA
                #     self.params_to_write['vocab_reverse'] = vocab_reverse
                else:
                    self.params_to_write[deepnovo_param] = param_value

        self.write_params_file()

        self.params['command_list'] = [
            sys.executable,
            self.exe,
            '--{0}'.format(self.params['translations']['deepnovo_mode'])
        ]
        print('executing command list')
        print(' '.join(self.params['command_list']))

        return self.params


    def postflight(self):
        '''
        Reformats the DeepNovo output file
        '''
        deepnovo_header = [
            'scan',
            'predicted_sequence',
            'predicted_score',
            'predicted_position_score',
        ]

        mod_lookup = {
            'Cmod': ('C', 'Carbamidomethyl'),
            'Mmod': ('M', 'Oxidation'),
            'Nmod': ('N', 'Deamidated'),
            'Qmod': ('Q', 'Deamidated'),
        }

        translated_headers = []
        header_translations = self.UNODE_UPARAMS[
            'header_translations']['uvalue_style_translation']
        for original_header_key in deepnovo_header:
            ursgal_header_key = header_translations[original_header_key]
            if ursgal_header_key not in translated_headers:
                translated_headers.append(ursgal_header_key)

        translated_headers += [
            'Raw data location',
            'Spectrum Title',
            'Charge',
            'Retention Time (s)',
            'Exp m/z',
            'Modifications',
        ]

        csv_kwargs = {}
        if sys.platform == 'win32':
            csv_kwargs['lineterminator'] = '\n'
        else:
            csv_kwargs['lineterminator'] = '\r\n'
        csv_writer = csv.DictWriter(
            open(self.params['translations']['output_file_incl_path'], 'w'),
            fieldnames=translated_headers,
            **csv_kwargs
        )

        deepnovo_output = os.path.join(
            self.params['translations']['tmp_output_file_incl_path']
        )
        csv_reader = csv.DictReader(
            open(deepnovo_output, 'r'),
            fieldnames=translated_headers,
            delimiter='\t',
        )

        csv_writer.writeheader()
        for n, line_dict in enumerate(csv_reader):
            if n == 0:
                continue
            line_dict['Raw data location'] = os.path.abspath(
                self.params['translations']['mgf_input_file']
            )
            line_dict['Charge'] = self.scan_lookup[line_dict['Spectrum ID']]

        #     ############################################
        #     # all fixing here has to go into unify csv! #
        #     ############################################

            tmp_seq = ''
            seq_list = line_dict['Sequence'].split(',')
            tmp_mods = []
            for n, aa in enumerate(seq_list):
                if 'mod' in aa:
                    org_aa, mod = mod_lookup[aa]
                    tmp_mods.append(
                        '{0}:{1}'.format(mod, n+1)
                    )
                    tmp_seq += org_aa
                else:
                    tmp_seq += aa
            line_dict['Sequence'] = tmp_seq
            line_dict['Modifications'] = ';'.join(tmp_mods)
            if line_dict['Sequence'] == '':
                continue
            csv_writer.writerow(line_dict)
        return

    def write_params_file(self):
        with open(self.params['translations']['params_file'], 'w') as io:
            print('''
# Copyright 2017 Hieu Tran. All Rights Reserved.
#
# DeepNovo is publicly available for non-commercial uses.
# ==============================================================================

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf


# ==============================================================================
# FLAGS (options) for this app
# ==============================================================================


tf.app.flags.DEFINE_string("train_dir", # flag_name
                           r"{train_dir}", # default_value
                           "Training directory.") # docstring

tf.app.flags.DEFINE_integer("direction",
                            {direction},
                            "Set to 0/1/2 for Forward/Backward/Bi-directional.")

tf.app.flags.DEFINE_boolean("use_intensity",
                            {use_intensity},
                            "Set to True to use intensity-model.")

tf.app.flags.DEFINE_boolean("shared",
                            {shared},
                            "Set to True to use shared weights.")

tf.app.flags.DEFINE_boolean("use_lstm",
                            {use_lstm},
                            "Set to True to use lstm-model.")

tf.app.flags.DEFINE_boolean("knapsack_build",
                            {knapsack_build},
                            "Set to True to build knapsack matrix.")

tf.app.flags.DEFINE_boolean("train",
                            False,
                            "Set to True for training.")

tf.app.flags.DEFINE_boolean("test_true_feeding",
                            False,
                            "Set to True for testing.")

tf.app.flags.DEFINE_boolean("decode",
                            False,
                            "Set to True for decoding.")

tf.app.flags.DEFINE_boolean("beam_search",
                            {beam_search},
                            "Set to True for beam search.")

tf.app.flags.DEFINE_integer("beam_size",
                            {beam_size},
                            "Number of optimal paths to search during decoding.")

tf.app.flags.DEFINE_boolean("search_db",
                            False,
                            "Set to True to do a database search.")

tf.app.flags.DEFINE_boolean("search_denovo",
                            False,
                            "Set to True to do a denovo search.")

tf.app.flags.DEFINE_boolean("search_hybrid",
                            False,
                            "Set to True to do a hybrid, db+denovo, search.")

tf.app.flags.DEFINE_boolean("test",
                            False,
                            "Set to True to test the prediction accuracy.")

tf.app.flags.DEFINE_boolean("header_seq",
                            False,
                            "Set to False if peptide sequence is not provided.")

tf.app.flags.DEFINE_boolean("decoy",
                            False,
                            "Set to True to search decoy database.")

FLAGS = tf.app.flags.FLAGS


# ==============================================================================
# GLOBAL VARIABLES for VOCABULARY
# ==============================================================================


# Special vocabulary symbols - we always put them at the start.
_PAD = '_PAD'
_GO = '_GO'
_EOS = '_EOS'
_START_VOCAB = [_PAD, _GO, _EOS]

PAD_ID = 0
GO_ID = 1
EOS_ID = 2

vocab_reverse = ['A',
                 'R',
                 'N',
                 'Nmod',
                 'D',
                 #~ 'C',
                 'Cmod',
                 'E',
                 'Q',
                 'Qmod',
                 'G',
                 'H',
                 'I',
                 'L',
                 'K',
                 'M',
                 'Mmod',
                 'F',
                 'P',
                 'S',
                 'T',
                 'W',
                 'Y',
                 'V',
                ]

vocab_reverse = _START_VOCAB + vocab_reverse
print("vocab_reverse ", vocab_reverse)

vocab = dict([(x, y) for (y, x) in enumerate(vocab_reverse)])
print("vocab ", vocab)

vocab_size = len(vocab_reverse)
print("vocab_size ", vocab_size)


# ==============================================================================
# GLOBAL VARIABLES for THEORETICAL MASS
# ==============================================================================


mass_H = 1.0078
mass_H2O = 18.0106
mass_NH3 = 17.0265
mass_N_terminus = 1.0078
mass_C_terminus = 17.0027
mass_CO = 27.9949

mass_AA = dict([
           ('_PAD', 0.0),
           ('_GO', mass_N_terminus-mass_H),
           ('_EOS', mass_C_terminus+mass_H),
           ('A', 71.03711),
           ('R', 156.10111),
           ('N', 114.04293),
           ('Nmod', 115.02695),
           ('D', 115.02694),
           # ('C', 103.00919),
           ('Cmod', 160.03065),
           # ('Cmod', 161.01919),
           ('E', 129.04259),
           ('Q', 128.05858),
           ('Qmod', 129.0426),
           ('G', 57.02146),
           ('H', 137.05891),
           ('I', 113.08406),
           ('L', 113.08406),
           ('K', 128.09496),
           ('M', 131.04049),
           ('Mmod', 147.0354),
           ('F', 147.06841),
           ('P', 97.05276),
           ('S', 87.03203),
           ('T', 101.04768),
           ('W', 186.07931),
           ('Y', 163.06333),
           ('V', 99.06841),
          ])

mass_ID = [mass_AA[vocab_reverse[x]] for x in range(vocab_size)]
mass_ID_np = np.array(mass_ID, dtype=np.float32)

mass_AA_min = mass_AA["G"] # 57.02146


# ==============================================================================
# GLOBAL VARIABLES for PRECISION, RESOLUTION, temp-Limits of MASS & LEN
# ==============================================================================


# if change, need to re-compile cython_speedup
SPECTRUM_RESOLUTION = 10 # bins for 1.0 Da = precision 0.1 Da
#~ SPECTRUM_RESOLUTION = 20 # bins for 1.0 Da = precision 0.05 Da
#~ SPECTRUM_RESOLUTION = 40 # bins for 1.0 Da = precision 0.025 Da
#~ SPECTRUM_RESOLUTION = 50 # bins for 1.0 Da = precision 0.02 Da
#~ SPECTRUM_RESOLUTION = 80 # bins for 1.0 Da = precision 0.0125 Da
print("SPECTRUM_RESOLUTION ", SPECTRUM_RESOLUTION)

# if change, need to re-compile cython_speedup
WINDOW_SIZE = 10 # 10 bins
print("WINDOW_SIZE ", WINDOW_SIZE)

MZ_MAX = 3000.0
MZ_SIZE = int(MZ_MAX * SPECTRUM_RESOLUTION) # 30k

KNAPSACK_AA_RESOLUTION = 10000 # 0.0001 Da
mass_AA_min_round = int(round(mass_AA_min * KNAPSACK_AA_RESOLUTION)) # 57.02146
KNAPSACK_MASS_PRECISION_TOLERANCE = 100 # 0.01 Da
num_position = 0

PRECURSOR_MASS_PRECISION_TOLERANCE = 0.01

# ONLY for accuracy evaluation
#~ PRECURSOR_MASS_PRECISION_INPUT_FILTER = 0.01
PRECURSOR_MASS_PRECISION_INPUT_FILTER = 1000
AA_MATCH_PRECISION = 0.1

# skip (x > MZ_MAX,MAX_LEN)
MAX_LEN = 50 if FLAGS.decode else 30
print("MAX_LEN ", MAX_LEN)

# We use a number of buckets and pad to the closest one for efficiency.
_buckets = [12, 22, 32]
#~ _buckets = [12,22,32,42,52] # for decode
print("_buckets ", _buckets)


# ==============================================================================
# HYPER-PARAMETERS of the NEURAL NETWORKS
# ==============================================================================


num_ion = 8 # 2
print("num_ion ", num_ion)

l2_loss_weight = 0.0 # 0.0
print("l2_loss_weight ", l2_loss_weight)

#~ encoding_cnn_size = 4 * (RESOLUTION//10) # 4 # proportion to RESOLUTION
#~ encoding_cnn_filter = 4
#~ print("encoding_cnn_size ", encoding_cnn_size)
#~ print("encoding_cnn_filter ", encoding_cnn_filter)

embedding_size = 512
print("embedding_size ", embedding_size)

num_layers = 1
num_units = 512
print("num_layers ", num_layers)
print("num_units ", num_units)

keep_conv = 0.75
keep_dense = 0.5
print("keep_conv ", keep_conv)
print("keep_dense ", keep_dense)

batch_size = 128
print("batch_size ", batch_size)

epoch_stop = 20 # 50
print("epoch_stop ", epoch_stop)

train_stack_size = 4500
valid_stack_size = 15000 # 10%
test_stack_size = 4000
buffer_size = {buffer_size}
print("train_stack_size ", train_stack_size)
print("valid_stack_size ", valid_stack_size)
print("test_stack_size ", test_stack_size)
print("buffer_size ", buffer_size)

steps_per_checkpoint = 100 # 20 # 100 # 2 # 4 # 200
random_test_batches = 10
print("steps_per_checkpoint ", steps_per_checkpoint)
print("random_test_batches ", random_test_batches)

max_gradient_norm = 5.0
print("max_gradient_norm ", max_gradient_norm)


# ==============================================================================
# DATASETS
# ==============================================================================

# YEAST-LOW-COON_2013-PEAKS-DB-DUP
data_format = "mgf"
cleavage_rule = "{cleavage_rule}"
num_missed_cleavage = {num_missed_cleavage}
fixed_mod_list = ['C']
var_mod_list = ['N', 'Q', 'M']
precursor_mass_tolerance = {precursor_mass_tolerance_da} # Da
precursor_mass_ppm = {precursor_mass_tolerance_ppm}/1000000 # ppm (20 better) # instead of absolute 0.01 Da
knapsack_file = r"{knapsack_file}"
# training/testing/decoding files
# input_file_train = r"data.training/dia.xchen.nov27/fraction_1.mgf.split.train.dup"
# input_file_valid = r"data.training/dia.xchen.nov27/fraction_1.mgf.split.valid.dup"
# input_file_test = r"C:\\Users\\Admin\\Desktop\\ursgal_dev\\ursgal\\ursgal\\resources\\platform_independent\\arc_independent\\deepnovo_0_0_1\\data.training\\yeast.low.coon_2013\\peaks.db.mgf.test.dup"
decode_test_file = r"{mgf_new_input_file}"
decode_output_file = r"{output_path}"
# denovo files
denovo_input_file = r"{mgf_new_input_file}"
denovo_output_file = r"{output_path}"
# db files
db_fasta_file = r"{db_fasta_file}"
db_input_file = r"{mgf_new_input_file}"
db_output_file = r"{output_path}"
if FLAGS.decoy:
  db_output_file += ".decoy"
# hybrid files
hybrid_input_file = r"{mgf_new_input_file}"
hybrid_denovo_file = hybrid_input_file + ".deepnovo_hybrid_denovo"
hybrid_output_file = r"{output_path}"
# if FLAGS.decoy:
#   hybrid_output_file += ".decoy"
# test accuracy
# predicted_format = "deepnovo"
# target_file = "data.training/dia.xchen.nov27/fraction_1.mgf.split.test.dup.target"
# predicted_file = denovo_output_file
# accuracy_file = predicted_file + ".accuracy"
# ==============================================================================
'''.format(
                **self.params_to_write),
                file=io
            )
        return



#!/usr/bin/env python3
# encoding: utf-8
'''

Test the filter_csv function for xtandem engine

'''
import ursgal
import csv
import os


R = ursgal.UController()


TESTS = [
    {
        'modifications' : [ 'M,opt,any,Oxidation' ],
        'type' : 'opt',
        'result_dict': {
            '_id'   : 0,
            'aa'    : 'M',
            'mass'  : 15.994915,
            'pos'   : 'any',
            'name'  : 'Oxidation',
            'composition' : {'O': 1},
            'org'   : 'M,opt,any,Oxidation',
            'id'    : '35',
            'unimod': True,
        },
    },
    {
        'modifications' : [ 'C,fix,any,Carbamidomethyl' ],
        'type' : 'fix',
        'result_dict': {
            '_id'   : 0,
            'aa'    : 'C',
            'mass'  : 57.021464,
            'pos'   : 'any',
            'name'  : 'Carbamidomethyl',
            'composition' : {'O': 1, 'N': 1, 'C': 2, 'H': 3},
            'org'   : 'C,fix,any,Carbamidomethyl',
            'id'    : '4',
            'unimod': True,
        },
    },
]

def map_mods_test():
    for test_id, test_dict in enumerate(TESTS):
        yield map_mods, test_dict

def map_mods( test_dict ):
    R = ursgal.UController(
        params = {'modifications' : test_dict['modifications']}
        )
    ursgal.UNode.map_mods(R)
    map_mod_dict = R.params[ 'mods' ][test_dict['type']][0]
    for k, v in test_dict['result_dict'].items():
        assert v == map_mod_dict[k]
    # os.remove(output_csv)

if __name__ == '__main__':
    print(__doc__)
    for test_id, test_dict in enumerate(TESTS):
        map_mods(test_dict)



#!/usr/bin/env python
import ursgal
import os

class get_ftp_files_1_0_0( ursgal.UNode ):
    """
    get_ftp_files_1_0_0 UNode

    Downloads files from FTP servers
    
    Note:
        meta info param 'output_extensions' is by default txt, so that the 
        temporary txt json files get properly deleted

    Args:
        * ftp_url            
        * folder             
        * login              
        * password           
        * include_ext        
        * output_folder      
        * max_number_of_files
        * blocksize          
    
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Get FTP Files',
        'version'            : '1.0.0',
        'release_date'       : '2016-3-4',
        'engine_type' : {
            'fetcher' : True,
        },
        'input_extensions'   : [],
        'output_extensions'  : ['.txt'],
        'output_suffix'      : None,
        'in_development'     : False,
        'include_in_git'     : True,
        'distributable'      : True,
        'utranslation_style' : 'get_ftp_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'get_ftp_files_1_0_0.py',
                },
            },
        },
        'citation' : \
            'Kremer, L. P. M., Leufken, J., Oyunchimeg, P., Schulze, S. & '\
            'Fufezan, C. (2016) Ursgal, Universal Python Module Combining '\
            'Common Bottom-Up Proteomics Tools for Large-Scale Analysis. J. '\
            'Proteome res. 15, 788-794.',
    }

    def __init__(self, *args, **kwargs):
        super(get_ftp_files_1_0_0, self).__init__(*args, **kwargs)

    def _execute( self ):
        '''
        Downloads files from FTP server

        '''
        print('[ -ENGINE- ] Executing FTP Download ..')
        # self.time_point(tag = 'execution')
        main = self.import_engine_as_python_function()
        downloaded_files = main(
            ftp_url             = self.params.get('ftp_url', None),
            folder              = self.params.get('ftp_folder', None),
            login               = self.params.get('ftp_login', None),
            password            = self.params.get('ftp_password', None),
            include_ext         = self.params.get('ftp_include_ext', None),
            output_folder       = self.params.get('ftp_output_folder', None),
            max_number_of_files = self.params.get('ftp_max_number_of_files', None),
            blocksize           = self.params.get('ftp_blocksize', None)
        )
        # self.print_execution_time(tag='execution')
        self.io['output']['finfo']['dir'] =  os.path.dirname( downloaded_files[-1] )
        self.io['output']['finfo']['file'] = os.path.basename( downloaded_files[-1] )
        return



#!/usr/bin/env python
'''
Unifies the result csvs

usage:
    ./unify_csv_1_0_0.py <input_file> <output_file> <uberSearch_lookup.pkl> <search_engine> <score_colname>

Fixes are listed in the main function. Resulting csv have unified fields, which
is most important for consitent modification formattting.
'''

from __future__ import print_function
import sys
import os
import pickle
import csv
import ursgal
import pyqms
import pprint
# import ursgal.ukb
import re
from collections import defaultdict
from copy import deepcopy as dc
import itertools
from decimal import *

# import time
# increase the field size limit to avoid crash if protein merge tags
# become too long does not work under windows
if sys.platform != 'win32':
    csv.field_size_limit(sys.maxsize)


DIFFERENCE_14N_15N = ursgal.ukb.DIFFERENCE_14N_15N
MOD_POS_PATTERN = re.compile(r'(?P<modname>.*):(?P<pos>[0-9]*)$')


def main(input_file=None, output_file=None, scan_rt_lookup=None,
         params=None, search_engine=None, score_colname=None):
    '''
    Arguments:
        input_file (str): input filename of csv which should be unified
        output_file (str): output filename of csv after unifying
        scan_rt_lookup (dict): dictionary with entries of scanID to
            retention time under key 'scan_2_rt'
        force (bool): force True or False
        params (dict): params as passed by ursgal
        search_engine(str): the search engine the csv file stems from
        score_colname (str): the column names of the search engine's
            score (i.e. 'OMSSA:pvalue')

    List of fixes

    All engines
        * Retention Time (s) is correctly set using _ursgal_lookup.pkl
          During mzML conversion to mgf the retention time for every spec
          is stored in a internal lookup and used later for setting the RT.
        * All modifications are checked if they were given in
          params['modifications'], converted to the name that was given
          there and sorted according to their position.
        * Fixed modifications are added in 'Modifications', if not reported
          by the engine.
        * The monoisotopic m/z for for each line is calculated (uCalc m/z),
          since not all engines report the monoisotopic m/z
        * Mass accuracy calculation (in ppm), also taking into account that
          not always the monoisotopic peak is picked
        * Rows describing the same PSM (i.e. when two proteins share the
          same peptide) are merged to one row.

    X!Tandem
        * 'RTINSECONDS=' is stripped from Spectrum Title if present in .mgf or
          in search result.

    Myrimatch
        * Spectrum Title is corrected
        * 15N label is not formatted correctly these modifications are
          removed for further analysis.
        * When using 15N modifications on amino acids and Carbamidomethyl
          myrimatch reports sometimes Carboxymethylation on Cystein.

    MS-GF+
        * 15N label is not formatted correctly these modifications are
          removed for further analysis.
        * 'Is decoy' column is properly set to true/false
        * Carbamidomethyl is updated and set if label is 15N

    OMSSA
        * Carbamidomethyl is updated and set

    MS-Amanda
        * multiple protein ID per peptide are splitted in two entries.
          (is done in MS-Amanda postflight)

    MSFragger
        * 15N modification have to be removed from Modifications and the
          merged modifications have to be corrected.

    pGlyco
        * reformat modifications
        * reformat glycan
    '''
    print(
        '''
[ unifycsv ] Converting {0} of engine {1} to unified CSV format...
        '''.format(
            os.path.basename(input_file),
            search_engine,
        )
    )

    # get the rows which define a unique PSM (i.e. sequence+spec+score...)
    # psm_defining_colnames = get_psm_defining_colnames(score_colname, search_engine)
    joinchar              = params['translations']['protein_delimiter']
    do_not_delete         = False
    created_tmp_files     = []
    use15N                = False
    search_engine         = search_engine.lower()

    if 'label' in params.keys():
        if params['label'] == '15N':
            use15N = True
    else:
        params['label'] = '14N'

    pyqms_mz_calc = params['translations']['use_pyqms_for_mz_calculation']
    # print(use15N)
    # sys.exit(1)
    # aa_exception_dict = params['translations']['aa_exception_dict']
    n_term_replacement = {
        'Ammonia-loss' : None,
        'Trimethyl'    : None,
        'Gly->Val'     : None,
    }
    fixed_mods   = {}
    opt_mods     = {}
    mod_dict     = {}
    cam          = False

    # modification masses are rounded to allow matching to unimod
    no_decimals = params['translations']['rounded_mass_decimals']
    if 'pipi' in search_engine:
        no_decimals = 1
    if 'moda' in search_engine:
        no_decimals = 0
    mass_format_string = '{{0:3.{0}f}}'.format(no_decimals)

    # mod pattern
    mod_pattern = re.compile( r''':(?P<pos>[0-9]*$)''' )

    for mod_type in ['fix', 'opt']:
        for modification in params['mods'][mod_type]:
            aa = modification['aa']
            pos = modification['pos']
            name = modification['name']
            if name not in mod_dict.keys():
                mod_dict[name] = {
                    'mass' : modification['mass'],
                    'aa' : set(),
                    'pos': set(),
                }
            mod_dict[name]['aa'].add(aa)
            mod_dict[name]['aa'].add(pos)
            if 'N-term' in pos:
                n_term_replacement[name] = aa
            if mod_type == 'fix':
                fixed_mods[aa] = name
                if aa == 'C' and name == 'Carbamidomethyl':
                    cam = True
                    # allow also Carbamidomnethyl on U, since the mod name gets changed
                    # already in upeptide_mapper
                    # According to unimod, the mnodification is also on Selenocystein
                    # otherwise we should change that back so that it is skipped...
                    mod_dict['Carbamidomethyl']['aa'].add('U')
                    fixed_mods['U'] = 'Carbamidomethyl'
            if mod_type == 'opt':
                opt_mods[aa] = name

    if 'msfragger' in search_engine:
        ##########################
        # msfragger mod merge block
        # calculate possbile mod combos...
        # if 15N add artifical mods...
        getcontext().prec = 8
        getcontext().rounding = ROUND_UP
        # mod_dict_list = params['mods']['opt'] + params['mods']['fix']
        if use15N:
            aminoacids_2_check = set()
            for modname in mod_dict.keys():
                aminoacids_2_check |= mod_dict[modname]['aa']
            additional_15N_modifications = []
            for aminoacid, N15_Diff in ursgal.ukb.DICT_15N_DIFF.items():
                if aminoacid not in aminoacids_2_check:
                    continue
                if '_15N_{0}'.format(aminoacid) in mod_dict.keys():
                    print('''
                        Error in unify_csv
                        New mod_name already present in mod_dict'
                        This should never happen'''
                    )
                    sys.exit(1)
                mod_dict['_15N_{0}'.format(aminoacid)] = {
                    'mass' : N15_Diff,
                    'aa' : set([aminoacid]),
                    'pos': set(['any']),
                }
                # additional_dict = {
                #     'name' : '_15N_{0}'.format(aminoacid),
                #     'mass' : N15_Diff,
                #     'aa'   : aminoacid,
                #     'pos'  : 'any'

                # }
            #     additional_15N_modifications.append(
            #         additional_dict
            #     )
            # mod_dict_list += additional_15N_modifications

        # mod_lookup = {} #d['name'] for d in self.params['mods']['opt']]
        # for mod_dict in mod_dict_list:
        #     if mod_dict['name'] not in mod_lookup.keys():
        #         mod_lookup[mod_dict['name']] = []
        #     mod_lookup[ mod_dict['name'] ].append(mod_dict)

        mod_names = []
        for mod in sorted(list(mod_dict.keys())):
            mod_names.extend(itertools.repeat(mod, len(mod_dict[mod]['aa'])))
        mass_to_mod_combo = {}
        # we cover all combocs of mods
        for iter_length in range(2, len(mod_names) + 1):
            for name_combo in itertools.combinations(mod_names, iter_length):
                mass = 0
                for name in name_combo:
                    mass += Decimal(mod_dict[name]['mass'])
                rounded_mass = mass_format_string.format(mass)
                if rounded_mass not in mass_to_mod_combo.keys():
                    mass_to_mod_combo[rounded_mass] = set()
                mass_to_mod_combo[ rounded_mass ].add( name_combo )
        # print(mass_to_mod_combo.keys())
        # sys.exit(1)
        #msfragger mod merge block end
        ##############################

    aa_exception_dict = params['translations']['aa_exception_dict']
    for unusual_aa, original_aa_dict in aa_exception_dict.items():
        if 'unimod_name' in original_aa_dict.keys():
            if original_aa_dict['unimod_name'] not in mod_dict.keys():
                mod_dict[original_aa_dict['unimod_name']] = {
                    'mass' : 0.0,
                    'aa' : set(),
                    'pos': set(),
                }
            mod_dict[original_aa_dict['unimod_name']]['aa'].update(original_aa_dict['original_aa'])

    cc = ursgal.ChemicalComposition()
    ursgal.GlobalUnimodMapper._reparseXML()
    de_novo_engines = [
        'novor',
        'pepnovo',
        'uninovo',
        'deepnovo',
        'unknown_engine'
    ]
    database_search_engines = [
        'msamanda',
        'msgf',
        'myrimatch',
        'omssa',
        'xtandem',
        'msfragger',
        'pglyco',
    ]
    open_mod_search_engines = [
        'pipi',
        'moda',
    ]
    de_novo = False
    database_search = False
    open_mod_search = False
    for de_novo_engine in de_novo_engines:
        if de_novo_engine in search_engine:
            de_novo = True
    for db_se in database_search_engines:
        if db_se in search_engine:
            database_search = True
    for om_se in open_mod_search_engines:
        if om_se in search_engine:
            open_mod_search = True

    if params['translations']['enzyme'] != 'nonspecific':
        allowed_aa, cleavage_site, inhibitor_aa = params['translations']['enzyme'].split(';')
    else:
        allowed_aa    = ''.join( list( ursgal.ukb.NITROGENS.keys() ) )
        cleavage_site = 'C'
        inhibitor_aa  = ''
    allowed_aa += '-'

    if database_search is True:
        non_enzymatic_peps = set()
        conflicting_uparams = defaultdict(set)
        fasta_lookup_name = os.path.basename(
            os.path.abspath(
                params['translations']['database']
            )
        )
        upeptide_map_sort_key   = 'Protein ID'
        upeptide_map_other_keys = [
            'Sequence Start',
            'Sequence Stop',
            'Sequence Pre AA',
            'Sequence Post AA',
        ]
    # psm_counter = Counter()
    # if a PSM with multiple rows is found (i.e. in omssa results), the psm
    # rows are merged afterwards

    output_file_object = open(output_file, 'w')
    protein_id_output  = open(output_file + '_full_protein_names.txt', 'w')
    mz_buffer          = {}
    csv_kwargs         = {
        'extrasaction' : 'ignore'
    }
    if sys.platform == 'win32':
        csv_kwargs['lineterminator'] = '\n'
    else:
        csv_kwargs['lineterminator'] = '\r\n'

    ze_only_buffer = {}

    app_mass_to_name_list_buffer = {}

    with open( input_file, 'r' ) as in_file:
        csv_input  = csv.DictReader(
            in_file
        )
        # recheck if fieldnames are correct. These are corrected in the upeptide
        # mapper but if the search engine is a de novo engine then the fields
        # might be incorrect
        output_fieldnames = list(csv_input.fieldnames)
        for remove_fieldname in [
            'proteinacc_start_stop_pre_post_;',
            'Start',
            'Stop',
            'NIST score',
            'gi',
            'Accession',
        ]:
            if remove_fieldname not in output_fieldnames:
                continue
            output_fieldnames.remove(remove_fieldname)
        new_fieldnames = [
            'uCalc m/z',
            'Accuracy (ppm)',
            'Mass Difference',
            'Protein ID',
            'Sequence Start',
            'Sequence Stop',
            'Sequence Pre AA',
            'Sequence Post AA',
            'Complies search criteria',
            'Conflicting uparam',
            'Search Engine',
        ]

        for new_fieldname in new_fieldnames:
            if new_fieldname not in output_fieldnames:
                output_fieldnames.insert( -5, new_fieldname )
        csv_output = csv.DictWriter(
            output_file_object,
            output_fieldnames,
            **csv_kwargs
        )
        csv_output.writeheader()
        print('''[ unify_cs ] Buffering csv file''')
        csv_file_buffer = []
        for line_dict in csv_input:
            csv_file_buffer.append(line_dict)
        total_lines = len(csv_file_buffer)
        print('''[ unify_cs ] Buffering csv file done''')
        all_molecules = set()
        all_charges = set()
        line_dict_collector = []
        for line_nr, line_dict in enumerate(csv_file_buffer):
            if line_nr % 500 == 0:
                print(
                    '[ unify_cs ] Processing line number: {0}/{1} '.format(
                        line_nr,
                        total_lines,
                    ),
                    end = '\r'
                )

            line_dict['Search Engine'] = search_engine
            ##########################
            # Spectrum Title block
            # reformatting Spectrum Title,
            if line_dict['Spectrum Title'] != '':
                '''
                Valid for:
                    OMSSA
                    MSGF+
                    X!Tandem
                    pGlyco
                '''
                if 'RTINSECONDS=' in line_dict['Spectrum Title']:
                    line_2_split = line_dict['Spectrum Title'].split(' ')[0].strip()
                elif line_dict['Spectrum Title'].endswith('.dta'):
                    '.'.join(line_2_split = line_dict['Spectrum Title'].split('.')[:-2])
                else:
                    line_2_split = line_dict['Spectrum Title']
                line_dict['Spectrum Title'] = line_2_split

                input_file_basename, spectrum_id, _spectrum_id, charge = line_2_split.split('.')[:4]
                pure_input_file_name = ''

            elif 'scan=' in line_dict['Spectrum ID']:
                pure_input_file_name                = os.path.basename(
                    line_dict['Raw data location']
                )
                input_file_basename = pure_input_file_name.split(".")[0]
                # not using os.path.splitext because we could have multiple file
                # extensions (i.e. ".mzml.gz")

                '''
                Valid for:
                    myrimatch
                '''
                spectrum_id = line_dict['Spectrum ID'].split('=')[-1]
                line_dict['Spectrum Title'] = '{0}.{1}.{1}.{2}'.format(
                    input_file_basename,
                    spectrum_id,
                    line_dict['Charge']
                )

            elif line_dict['Spectrum Title'] == '':
                '''
                Valid for:
                    Novor
                    DeepNovo
                    MSFragger
                '''
                pure_input_file_name = os.path.basename(
                    line_dict['Raw data location']
                )
                input_file_basename = pure_input_file_name.split(".")[0]
                spectrum_id = line_dict['Spectrum ID']
                line_dict['Spectrum Title'] = '{0}.{1}.{1}.{2}'.format(
                    input_file_basename,
                    spectrum_id,
                    line_dict['Charge']
                )
            else:
                raise Exception( 'New csv format present for engine {0}'.format( engine ) )

            #update spectrum ID from block above
            line_dict['Spectrum ID'] = spectrum_id

            # now check for the basename in the scan rt lookup
            # possible cases:
            #   - input_file_basename
            #   - input_file_basename + prefix
            #   - input_file_basename - prefix

            input_file_basename_for_rt_lookup = None
            if input_file_basename in scan_rt_lookup.keys():
                input_file_basename_for_rt_lookup = input_file_basename
            else:
                basename_with_prefix = '{0}_{1}'.format(
                    params['prefix'],
                    input_file_basename
                )
                basename_without_prefix  = input_file_basename.replace(
                    params['prefix'],
                    ''
                )
                if basename_with_prefix in scan_rt_lookup.keys():
                    input_file_basename_for_rt_lookup = basename_with_prefix
                elif basename_without_prefix in scan_rt_lookup.keys():
                    input_file_basename_for_rt_lookup = basename_without_prefix
                else:
                    print(
                        '''
[ WARNING ] Could not find scan ID {0} in scan_rt_lookup[ {1} ]
                        '''.format(
                            spectrum_id,
                            input_file_basename
                        )
                    )

            #END Spectrum Title block
            ##########################
            spectrum_id = int(spectrum_id)
            retention_time_in_minutes = \
                scan_rt_lookup[ input_file_basename_for_rt_lookup ][ 'scan_2_rt' ]\
                    [ spectrum_id ]

            # We should check if data has minute format or second format...
            if scan_rt_lookup[ input_file_basename_for_rt_lookup ]['unit'] == 'second':
                rt_corr_factor = 1
            else:
                rt_corr_factor = 60
            line_dict['Retention Time (s)'] = float( retention_time_in_minutes ) * rt_corr_factor

            # try:
            precursor_mz = scan_rt_lookup[ input_file_basename_for_rt_lookup ][
                'scan_2_mz' ][ spectrum_id ]
            # except:
            #     print('\n\n\n')
            #     print(input_file_basename_for_rt_lookup)
            #     print('spectrum_id', spectrum_id, type(spectrum_id))
            #     exit(1)
            line_dict['Exp m/z'] = round(precursor_mz, 10)

            #########################
            # Buffering corrections #
            #########################
            main_buffer_key = '{Sequence} || {Charge} || {Modifications} || '.format(
                **line_dict
            ) + params['label']
            if main_buffer_key not in ze_only_buffer.keys():
                line_dict_update = {}
                ######################
                # Modification block #
                ######################

                # check MSFragger crazy mod merge first...
                if 'msfragger' in search_engine:
                    # we have to reformat the modifications
                    # M|14$15.994915|17$57.021465 to 15.994915:14;57.021465:17
                    # reformat it in Xtandem style
                    ms_fragger_reformatted_mods = []
                    if line_dict['Modifications'] == 'M':
                        # M stand for Modifications here, not Methionine
                        line_dict['Modifications'] = ''
                    else:
                        mod_list = line_dict['Modifications']
                        for single_mod in mod_list.split('|'):
                            if single_mod in ['M','']:
                                continue
                            msfragger_pos, raw_msfragger_mass = single_mod.split('$')
                            msfragger_mass = mass_format_string.format(
                                # mass rounded as defined above
                                Decimal(raw_msfragger_mass)
                            )
                            msfragger_pos = int(msfragger_pos)
                            if msfragger_mass in mass_to_mod_combo.keys():
                                explainable_combos = []
                                for combo in mass_to_mod_combo[msfragger_mass]:
                                    combo_explainable = set([True])
                                    tmp_mods = []
                                    for new_name in combo:
                                        meta_mod_info = mod_dict[new_name]
                                        single_mod_check = set([True])
                                        '''
                                        meta_mod_info = {
                                            'aa': set of aa,
                                            'mass': 42.010565,
                                            'pos': set of pos,
                                        }
                                        '''
                                        #check aa
                                        if '*' not in meta_mod_info['aa'] and \
                                            line_dict['Sequence'][msfragger_pos] not in meta_mod_info['aa']:
                                                single_mod_check.add(False)
                                        # check pos
                                        if 'any' not in meta_mod_info['pos']:
                                            pos_to_check = set()
                                            if 'Prot-N-term' in meta_mod_info['pos'] or\
                                                'N-term' in meta_mod_info['pos']:
                                                pos_to_check.add(0)
                                            elif 'Prot-C-term' in meta_mod_info['pos'] or \
                                                'C-term' in meta_mod_info['pos']:
                                                pos_to_check.add(int(len(line_dict['Sequence'])) - 1)
                                            else:
                                                pass
                                            if pos_to_check != set():
                                                if msfragger_pos not in pos_to_check:
                                                    single_mod_check.add(False)

                                        if all(single_mod_check):
                                            # MS Frager starts counting at zero
                                            pos_in_peptide_for_format_str = msfragger_pos + 1
                                            # we keep mass here so that the
                                            # correct name is added later in already
                                            # existing code
                                            tmp_mods.append(
                                                '{0}:{1}'.format(
                                                    meta_mod_info['mass'],
                                                    pos_in_peptide_for_format_str
                                                )
                                            )
                                        else:
                                            combo_explainable.add(False)
                                    if all(combo_explainable):
                                        explainable_combos.append(tmp_mods)
                                if len(explainable_combos) > 1:
                                    print(
                                        '''
                                        [ WARNING ] Multiple modification combinations possible
                                        [ WARNING ] to explain reported modification mass
                                        [ WARNING ] The following combination was chosen to continue:
                                        [ WARNING ] {0}
                                        '''.format(
                                            sorted(explainable_combos)[0],
                                        )
                                    )
                                    # pprint.pprint(explainable_combos)
                                    # ms_fragger_reformatted_mods += sorted(explainable_combos)[0]
                                    # sys.exit(1)
                                elif len(explainable_combos) == 1:
                                    ms_fragger_reformatted_mods += sorted(explainable_combos)[0]
                                else:
                                    # no combos explainable
                                    ms_fragger_reformatted_mods.append(
                                        '{0}:{1}'.format(
                                            raw_msfragger_mass,
                                            msfragger_pos + 1
                                        )
                                    )
                            else:
                                # MS Frager starts counting at zero
                                ms_fragger_reformatted_mods.append(
                                    '{0}:{1}'.format(
                                        raw_msfragger_mass,
                                        msfragger_pos + 1
                                    )
                                )
                        # print(line_dict['Modifications'])
                        # print(mass_to_mod_combo.keys())
                        # print(ms_fragger_reformatted_mods)
                        # sys.exit(1)
                        line_dict['Modifications'] = ';'.join( ms_fragger_reformatted_mods )

                ##################################################
                # Some engines do not report fixed modifications #
                ##################################################
                if fixed_mods != {}:
                    for pos, aminoacid in enumerate(line_dict['Sequence']):
                        if aminoacid in fixed_mods.keys():
                            name = fixed_mods[aminoacid]
                            tmp = '{0}:{1}'.format(
                                name,
                                pos + 1
                            )
                            append_mod = True
                            if tmp in line_dict['Modifications']:
                                # everything is ok :)
                                append_mod = False
                            elif ':{0}'.format(pos + 1) in line_dict['Modifications']:
                                # there is a mod at that position but is it the right one ?
                                for _m in line_dict['Modifications'].split(';'):
                                    match = MOD_POS_PATTERN.search(_m)
                                    if match is not None:
                                        mod = match.group('modname')
                                        pos = match.group('pos')
                                        try:
                                            if round(mod_dict[name]['mass'], 5) == round(float(mod), 5):
                                                append_mod = False
                                        except:
                                            pass

                            if append_mod:
                                tmp_mods = line_dict['Modifications'].split(';')
                                tmp_mods.append(tmp)
                                line_dict['Modifications'] = ';'.join( tmp_mods )
                ##################################################################
                # Myrimatch, MSGF+, and MSFragger can not handle 15N that easily #
                # Report all AAs moded with unknown modification                 #
                # Note: masses are checked below to avoid any mismatch           #
                ##################################################################
                if use15N:
                    if 'myrimatch' in search_engine or \
                            'msgfplus_v9979' in search_engine:
                        for p in range(1,len(line_dict['Sequence'])+1):
                                line_dict['Modifications'] = \
                                    line_dict['Modifications'].replace(
                                        'unknown modification:{0}'.format(p),
                                        '',
                                        1,
                                    )
                    if 'myrimatch' in search_engine:
                        if 'Carboxymethyl' in line_dict['Modifications'] and cam is True:
                            line_dict['Modifications'] = line_dict['Modifications'].replace(
                                'Carboxymethyl',
                                'Carbamidomethyl'
                            )
                        elif 'Delta:H(6)C(3)O(1)' in line_dict['Modifications']:
                            line_dict['Modifications'] = line_dict['Modifications'].replace(
                                'Delta:H(6)C(3)O(1)',
                                'Carbamidomethyl'
                            )

                tmp_mods = []
                tmp_mass_diff = []
                for modification in line_dict['Modifications'].split(';'):
                    raw_modification = modification
                    Nterm = False
                    Cterm = False
                    skip_mod = False
                    if modification == '' or modification == 'null':
                        continue
                    pos, mod = None, None
                    if 'pglyco' in search_engine:
                        try:
                            pos = int(modification.split(',')[0])
                            mod_pglyco = ','.join(modification.split(',')[1:])
                            mod = mod_pglyco.split('[')[0]
                        except:
                            match = mod_pattern.search( modification )
                            pos = int( match.group('pos') )
                            mod = modification[ :match.start() ]
                    else:
                        match = mod_pattern.search( modification )
                        pos = int( match.group('pos') )
                        mod = modification[ :match.start() ]
                    assert pos is not None, '''
                            The format of the modification {0}
                            is not recognized by ursgal'''.format(
                                modification
                            )
                    if pos <= 1:
                        Nterm = True
                        new_pos = 1
                    elif pos > len(line_dict['Sequence']):
                        Cterm = True
                        new_pos = len(line_dict['Sequence'])
                    else:
                        new_pos = pos
                    aa = line_dict['Sequence'][ new_pos - 1 ].upper()
                    # if aa in fixed_mods.keys():
                    #     fixed_mods[ aminoacid ]
                    #     # fixed mods are corrected/added already
                    #     continue
                    if mod in mod_dict.keys():
                        correct_mod = False
                        if aa in mod_dict[mod]['aa']:
                            # everything is ok
                            correct_mod = True
                        elif Nterm or Cterm:
                            if '*' in mod_dict[mod]['aa']:
                                correct_mod = True
                                # still is ok
                        assert correct_mod is True,'''
                                A modification was reported for an aminoacid for which it was not defined
                                unify_csv cannot deal with this, please check your parameters and engine output
                                reported modification: {0} on {1}
                                modifications in parameters: {2}
                                '''.format(
                                    mod,
                                    aa,
                                    params['mods']
                                )
                    elif 'unknown modification' == mod:
                        modification_known = False
                        if aa in opt_mods.keys():
                            # fixed mods are corrected/added already
                            modification = '{0}:{1}'.format(opt_mods[aa],new_pos)
                            modification_known = True
                        assert modification_known == True,'''
                                unify csv does not work for the given unknown modification for
                                {0} {1} aa: {2}
                                maybe an unknown modification with terminal position was given?
                                '''.format(
                                    line_dict['Sequence'], modification, aa
                                )
                    else:
                        float_mod = float(mod)
                        masses_2_test = [float_mod]
                        if use15N:
                            substract_15N_diff = False
                            if aa in fixed_mods.keys() and 'msgfplus' in search_engine and pos != 0:
                                substract_15N_diff = True
                            if 'msfragger' in search_engine and float_mod > 4:
                                # maximum 15N labeling is 3.988 Da (R)
                                substract_15N_diff = True
                            if substract_15N_diff:
                                masses_2_test.append(float_mod - ursgal.ukb.DICT_15N_DIFF[aa])
                        # try:
                        #works always but returns empty list...
                        name_list = []
                        for mass_2_test in masses_2_test:
                            mass_buffer_key = mass_format_string.format(mass_2_test)
                            #buffer increases speed massively...
                            if mass_buffer_key not in app_mass_to_name_list_buffer.keys():
                                app_mass_to_name_list_buffer[mass_buffer_key] = ursgal.GlobalUnimodMapper.appMass2name_list(
                                    float(mass_buffer_key),
                                    decimal_places = no_decimals
                                )
                            name_list += app_mass_to_name_list_buffer[mass_buffer_key]
                        # print(name_list)
                        # except:
                        #     print('''
                        #         A modification was reported that was not included in the search parameters
                        #         unify_csv cannot deal with this, please check your parameters and engine output
                        #         reported modification: {0}
                        #         modifications in parameters: {1}
                        #         '''.format(mod, params['translations']['modifications'])
                        #     )
                        #     raise Exception('unify_csv failed because a '\
                        #         'modification was reported that was not '\
                        #         'given in params: {0}'.format(modification)
                        #     )
                        mapped_mod = False
                        for name in name_list:
                            if name in mod_dict.keys():
                                if aa in mod_dict[name]['aa']:
                                    modification = '{0}:{1}'.format(name, new_pos)
                                    mapped_mod = True
                                elif Nterm and '*' in mod_dict[name]['aa']:
                                    modification = '{0}:{1}'.format(name, 0)
                                    mapped_mod = True
                                else:
                                    continue
                            elif use15N and name in [
                                'Label:15N(1)',
                                'Label:15N(2)',
                                'Label:15N(3)',
                                'Label:15N(4)'
                            ]:
                                mapped_mod = True
                                skip_mod = True
                                break

                        if open_mod_search is True and mapped_mod is False:
                            skip_mod = True
                            tmp_mass_diff.append('{0}:{1}'.format(mod, pos))
                            continue
                        assert mapped_mod is True, '''
                                A mass was reported that does not map on any unimod or userdefined modification
                                or the modified aminoacid is not the specified one
                                unify_csv cannot deal with this, please check your parameters and engine output
                                sequence: {4}
                                reported mass: {0}
                                maps on: {1}
                                reported modified aminoacid: {2}
                                modifications in parameters: {3}
                                '''.format(
                                    mod,
                                    name_list,
                                    aa,
                                    params['mods'],
                                    line_dict['Sequence']
                                )
                    if modification in tmp_mods:
                        add_n_term_mod = False
                        _mod, _pos = modification.split(':')
                        _aa = line_dict['Sequence'][int(_pos) - 1]
                        if _aa in mod_dict[_mod]['aa']:
                            if 'N-term' in mod_dict[_mod]['aa']:
                                add_n_term_mod = True
                            elif 'Prot-N-term' in mod_dict[_mod]['aa']:
                                add_n_term_mod = True
                        if add_n_term_mod:
                            modification = modification.replace(
                                '{0}:1'.format(_mod),
                                '{0}:0'.format(_mod)
                            )
                        # else:
                        #     skip_mod = True
                    if skip_mod is True:
                        continue
                    tmp_mods.append(modification)
                if 'msfragger' in search_engine:
                    org_mass_diff = line_dict['Mass Difference']
                    tmp_mass_diff.append('{0}:n'.format(org_mass_diff))
                if 'pglyco' in search_engine:
                    line_dict['Sequence'] = line_dict['Sequence'].replace('J', 'N')
                line_dict_update['Modifications'] = ';'.join(tmp_mods)
                line_dict_update['Mass Difference'] = ';'.join(tmp_mass_diff)
                #
                # ^^--------- REPLACED MODIFICATIONS! ---------------^
                #
                for unimod_name in n_term_replacement.keys():
                    if '{0}:1'.format(unimod_name) in line_dict_update['Modifications'].split(';'):
                        if unimod_name in mod_dict.keys():
                            aa = mod_dict[unimod_name]['aa']
                            if aa != ['*']:
                                if line_dict['Sequence'][0] in aa:
                                    continue
                        line_dict_update['Modifications'] = line_dict_update['Modifications'].replace(
                            '{0}:1'.format(unimod_name),
                            '{0}:0'.format(unimod_name)
                            )
                ##########################
                # Modification block end #
                ##########################

                line_dict_update['Sequence'] = line_dict['Sequence']
                #
                # ^^--------- REPLACED SEQUENCE! ---------------^
                #
                # remove the double ';''
                if line_dict_update['Modifications'] != '':
                    tmp = []
                    positions = set()
                    for e in line_dict_update['Modifications'].split(';'):
                        if e == '':
                            # that remove the doubles ....
                            continue
                        else:
                            # other way to do it...
                            # pos_of_split_point = re.search( ':\d*\Z', e )
                            # pattern = re.compile( r''':(?P<pos>[0-9]*$)''' )
                            for occ, match in enumerate(mod_pattern.finditer(e)):
                                mod = e[:match.start()]
                                mod_pos = e[match.start()+1:]
                                # mod, pos = e.split(':')
                                m = (int(mod_pos), mod)
                                if m not in tmp:
                                    tmp.append(m)
                                    positions.add(int(mod_pos))
                    tmp.sort()
                    line_dict_update['Modifications'] = ';'.join(
                        [
                            '{m}:{p}'.format( m=mod, p=pos) for pos, mod in tmp
                        ]
                    )
                    if len(tmp) != len(positions):
                        print(
                            '[ WARNING ] {Sequence}#{Modifications} will be skipped, because it contains two mods at the same position!'.format(
                                **line_dict_update
                            )
                        )
                        continue

                all_molecules.add(
                    '{Sequence}#{Modifications}'.format(
                        **line_dict_update
                    )
                )
                all_charges.add(int(line_dict['Charge']))


                # ------------
                # BUFFER END
                # -----------
                ze_only_buffer[ main_buffer_key ] = line_dict_update

            line_dict_update = ze_only_buffer[ main_buffer_key ]
            line_dict.update( line_dict_update )

            if 'pglyco' in search_engine:
                try:
                    Hex, HexNAc, NeuAc, NeuGc, dHex = line_dict[
                        'Glycan'].strip(' ').split(' ')
                except:
                    print(line_dict['Glycan'])
                    exit()
                line_dict['Glycan'] = 'Hex({0})HexNac({1})NeuAc({2})NeuGc({3})dHex({4})'.format(
                    Hex,
                    HexNAc,
                    NeuAc,
                    NeuGc,
                    dHex,
                )

            # protein block, only for database search engine
            if database_search is True:
                # check for correct cleavage sites and set a new field to
                # verify correct enzyme performance
                lookup_identifier = '{0}><{1}'.format(
                    line_dict['Sequence'],
                    fasta_lookup_name
                )
                if lookup_identifier not in conflicting_uparams.keys():
                    split_collector = { }
                    for key in [ upeptide_map_sort_key ] + upeptide_map_other_keys:
                        split_collector[ key ] = line_dict[key].split(joinchar)
                    sorted_upeptide_maps = []
                    for pos, protein_id in enumerate(sorted( split_collector[ upeptide_map_sort_key ] ) ):
                        dict_2_append = {
                            upeptide_map_sort_key : protein_id
                        }

                        for key in upeptide_map_other_keys:
                            dict_2_append[key] = split_collector[key][pos]
                        sorted_upeptide_maps.append(
                            dict_2_append
                        )
                    # pprint.pprint(line_dict)
                    # print(sorted_upeptide_maps)
                    # sys.exit(1)
                    if sorted_upeptide_maps == []:
                        print('''
[ WARNING ] The peptide {0} could not be mapped to the
[ WARNING ] given database {1}
[ WARNING ] {2}
[ WARNING ] This PSM will be skipped.
                            '''.format(
                                line_dict['Sequence'],
                                params['translations']['database'],
                            )
                        )
                        continue
                    peptide_fullfills_enzyme_specificity = False
                    last_protein_id = None
                    for major_protein_info_dict in sorted_upeptide_maps:
                        protein_specifically_cleaved   = False
                        nterm_correct = False
                        cterm_correct = False
                        '''
                            'Sequence Start',
                            'Sequence Stop',
                            'Sequence Pre AA',
                            'Sequence Post AA',

                        '''
                        protein_info_dict_buffer = []
                        if ';' in major_protein_info_dict['Sequence Start']:
                            tmp_split_collector =defaultdict(list)
                            for key in upeptide_map_other_keys:
                                tmp_split_collector[key] = major_protein_info_dict[key].split(';')
                            for pos, p_id in enumerate(tmp_split_collector['Sequence Start']):
                                dict_2_append = {
                                    'Protein ID' : major_protein_info_dict['Protein ID']
                                }
                                for key in tmp_split_collector.keys():
                                    dict_2_append[key] = tmp_split_collector[key][pos]
                                protein_info_dict_buffer.append(dict_2_append)
                        else:
                            protein_info_dict_buffer = [ major_protein_info_dict ]

                        for protein_info_dict in protein_info_dict_buffer:
                            if params['translations']['keep_asp_pro_broken_peps'] is True:
                                if line_dict['Sequence'][-1] == 'D' and\
                                        protein_info_dict['Sequence Post AA'] == 'P':
                                    cterm_correct = True
                                if line_dict['Sequence'][0] == 'P' and\
                                        protein_info_dict['Sequence Pre AA'] == 'D':
                                    nterm_correct = True

                            if cleavage_site == 'C':
                                if protein_info_dict['Sequence Pre AA'] in allowed_aa\
                                        or protein_info_dict['Sequence Start'] in ['1', '2', '3']:
                                    if line_dict['Sequence'][0] not in inhibitor_aa\
                                            or protein_info_dict['Sequence Start'] in ['1', '2', '3']:
                                        nterm_correct = True
                                if protein_info_dict['Sequence Post AA'] not in inhibitor_aa:
                                    if line_dict['Sequence'][-1] in allowed_aa\
                                         or protein_info_dict['Sequence Post AA'] == '-':
                                        cterm_correct = True

                            elif cleavage_site == 'N':
                                if protein_info_dict['Sequence Post AA'] in allowed_aa:
                                    if line_dict['Sequence'][-1] not in inhibitor_aa\
                                            or protein_info_dict['Sequence Post AA'] == '-':
                                        cterm_correct = True
                                if protein_info_dict['Sequence Pre AA'] not in inhibitor_aa\
                                    or protein_info_dict['Sequence Start'] in ['1', '2', '3']:
                                    if line_dict['Sequence'][0] in allowed_aa\
                                        or protein_info_dict['Sequence Start'] in ['1', '2', '3']:
                                        nterm_correct = True
                            # if line_dict['Sequence'] == 'SPRPGAAPGSR':
                            #     print(protein_info_dict)
                            #     print(nterm_correct, cterm_correct)
                            if params['translations']['semi_enzyme'] is True:
                                if cterm_correct is True or nterm_correct is True:
                                    protein_specifically_cleaved = True
                            elif cterm_correct is True and nterm_correct is True:
                                protein_specifically_cleaved = True
                            if protein_specifically_cleaved is True:
                                peptide_fullfills_enzyme_specificity = True
                                last_protein_id = protein_info_dict['Protein ID']
                    # we may test for further criteria to set this flag/fieldname
                    # e.g. the missed cleavage count etc.
                    if peptide_fullfills_enzyme_specificity is False:
                        non_enzymatic_peps.add( line_dict['Sequence'] )
                        conflicting_uparams[lookup_identifier].add(
                            'enzyme'
                        )

                    #check here if missed cleavage count is correct...
                    missed_cleavage_counter = 0
                    if params['translations']['enzyme'] != 'ACDEFGHIKLMNPQRSTVWY;C;':
                        for aa in allowed_aa:
                            if aa == '-':
                                continue
                            if cleavage_site == 'C':
                                if inhibitor_aa == '':
                                    missed_cleavage_pattern = aa
                                else:
                                    missed_cleavage_pattern = '{0}[^{1}]'.format(
                                        aa, inhibitor_aa
                                    )
                                missed_cleavage_counter += \
                                    len(re.findall(missed_cleavage_pattern, line_dict['Sequence']))
                            elif cleavage_site == 'N':
                                if inhibitor_aa == '':
                                    missed_cleavage_pattern = aa
                                else:
                                    missed_cleavage_pattern = '[^{1}]{0}'.format(
                                        aa, inhibitor_aa
                                    )
                                missed_cleavage_counter += \
                                    len(re.findall(missed_cleavage_pattern, line_dict['Sequence']))
                    if missed_cleavage_counter > params['translations']['max_missed_cleavages']:
                        conflicting_uparams[lookup_identifier].add('max_missed_cleavages')
                # count each PSM occurence to check whether row-merging is needed:
                # psm = tuple([line_dict[x] for x in psm_defining_colnames])
                # psm_counter[psm] += 1

                if len(conflicting_uparams[lookup_identifier]) == 0:
                    # all tested search criteria true
                    line_dict['Complies search criteria'] = 'true'
                else:
                    line_dict['Complies search criteria'] = 'false'
                    line_dict['Conflicting uparam'] = ';'.join(
                        sorted(conflicting_uparams[lookup_identifier])
                    )
            line_dict_collector.append(line_dict)

        # --------------------------------
        # mz calc and accuracy calc block
        # --------------------------------

        print('[ unify_cs ] calculating mz accuracies')
        # build IsotopologueLibrary
        molecule2hill_dict = {}
        for molecule in all_molecules:
            if 'X' in molecule.upper():
                cc.use(molecule.replace('X', ''))
            else:
                cc.use(molecule)
            if use15N:
                number_N = dc( cc['N'] )
                cc['15N'] = number_N
                del cc['N']
                if cam:
                    c_count = molecule.split('#')[0].count('C')
                    cc['14N'] = c_count
                    cc['15N'] -= c_count
            cc_hill = cc.hill_notation_unimod()
            molecule2hill_dict[molecule] = '+{0}'.format(cc_hill)

        if pyqms_mz_calc:
            isotopologue_dict = pyqms.IsotopologueLibrary(
                molecules=list(molecule2hill_dict.values()),
                charges=list(all_charges),
                verbose=True
            )

        # calculate m/z
        for collected_line_dict in line_dict_collector:
            molecule = '{Sequence}#{Modifications}'.format(
                **collected_line_dict
            )
            cc_hill = molecule2hill_dict[molecule].strip('+')
            charge = int(collected_line_dict['Charge'])
            if pyqms_mz_calc:
                isotopologue_mzs = isotopologue_dict[cc_hill][
                'env'][(('N', '0.000'),)][charge]['mz']

                calc_mz = round(isotopologue_mzs[0], 10)
                min_accuracy = (float(collected_line_dict['Exp m/z']) - calc_mz)/calc_mz * 1e6
                for iso_mz in isotopologue_mzs[1:]:
                    isotopologue_acc = (float(collected_line_dict['Exp m/z']) - iso_mz)/iso_mz * 1e6
                    if abs(isotopologue_acc) > abs(min_accuracy):
                        break
                    else:
                        min_accuracy = isotopologue_acc
            else:
                cc.use(molecule2hill_dict[molecule])
                mass = cc._mass()
                calc_mz = ursgal.ucore.calculate_mz(
                    mass,
                    int(collected_line_dict['Charge'])
                )
                min_accuracy = (float(collected_line_dict['Exp m/z']) - calc_mz)/calc_mz * 1e6
                for isotope in range(1,6):
                    iso_mz = ursgal.ucore.calculate_mz(
                        mass + isotope*1.008664904,
                        int(collected_line_dict['Charge'])
                    )
                    isotopologue_acc = (float(collected_line_dict['Exp m/z']) - iso_mz)/iso_mz * 1e6
                    if abs(isotopologue_acc) > abs(min_accuracy):
                        break
                    else:
                        min_accuracy = isotopologue_acc

            collected_line_dict['uCalc m/z'] = calc_mz
            if database_search is True:
                if 'Calc m/z' in collected_line_dict.keys() and\
                    collected_line_dict['Calc m/z'] == '':
                    collected_line_dict['Calc m/z'] = calc_mz

            collected_line_dict['Accuracy (ppm)'] = round(min_accuracy, 5)

            csv_output.writerow(collected_line_dict)

    output_file_object.close()

    if database_search is True:
        # upapa.purge_fasta_info( fasta_lookup_name )
        if len(non_enzymatic_peps) != 0:
            print(
                '''
                [ WARNING ] The following peptides do not reflect the enzyme
                [ WARNING ] specificity:
                [ WARNING ] {0}
                [ WARNING ] These PSMs are marked 'Complies search criteria' = 'False'
                '''.format(
                    non_enzymatic_peps
                )
            )
    # sys.exit(1)
    # if there are multiple rows for a PSM, we have to merge them aka rewrite the csv...
    # if psm_counter != Counter():
    #     if max(psm_counter.values()) > 1:
    #         merge_duplicate_psm_rows(output_file, psm_counter, psm_defining_colnames, params['translations']['psm_merge_delimiter'])
            '''
            to_be_written_csv_lines = merge_duplicate_psm_rows(
                to_be_written_csv_lines,
                psm_counter
            )
            '''
        '''
        do output_file magic with to_be_written_csv_lines
        '''
    if do_not_delete is False:
        created_tmp_files.append( output_file + '_full_protein_names.txt' )
    return created_tmp_files


# def get_psm_defining_colnames(score_colname, search_engine):
#     '''
#     Returns the all PSM-defining column names (i.e spectrum & peptide,
#     but also score field because sometimes the same PSMs are reported
#     with different scores...
#     '''
#     psm = [
#         'Spectrum Title',
#         'Sequence',
#         'Modifications',
#         'Charge',
#         'Is decoy',
#     ]
#     if 'msfragger' in search_engine:
#         psm.append('MSFragger:Neutral mass of peptide')
#     if score_colname:
#         psm.append(score_colname)
#     return psm


# def merge_rowdicts(list_of_rowdicts, joinchar, alt_joinchar='<|>'):
#     '''
#     Merges CSV rows. If the column values are conflicting, they
#     are joined with a character (joinchar).
#     Special case: proteinaccessions are not joined with the joinchar,
#     but rather with alt_joinchar.
#     '''
#     merged_d = {}
#     fieldnames = list_of_rowdicts[0].keys()
#     for fieldname in fieldnames:

#         joinchar_used = joinchar
#         # if fieldname == 'proteinacc_start_stop_pre_post_;':
#         #     joinchar_used = alt_joinchar

#         values = {d[fieldname] for d in list_of_rowdicts}
#         if len(values) == 1:
#             merged_d[fieldname] = list(values)[0]
#         else:
#             merged_d[fieldname] = joinchar_used.join(sorted(values))
#     return merged_d


# def merge_duplicate_psm_rows(unified_csv_path, psm_counter, psm_defining_colnames, joinchar):
#     '''
#     Rows describing the same PSM (i.e. when two proteins share the
#     same peptide) are merged to one row.
#     '''
#     rows_to_merge_dict = defaultdict(list)

#     tmp_file = unified_csv_path + ".tmp"
#     os.rename(unified_csv_path, tmp_file)
#     print('Merging rows of the same PSM...')
#     with open(tmp_file, 'r') as tmp, open(unified_csv_path, 'w', newline='') as out:
#         tmp_reader = csv.DictReader(tmp)
#         writer = csv.DictWriter(out, fieldnames=tmp_reader.fieldnames)
#         writer.writeheader()
#         for row in tmp_reader:
#             psm = tuple([row[x] for x in psm_defining_colnames])
#             # each unique combination of these should only have ONE row!
#             # i.e. combination of seq+spec+score
#             if psm_counter[psm] == 1:
#                 # no duplicate = no problem, we can just write the row again
#                 writer.writerow(row)
#             elif psm_counter[psm] > 1:
#                 # we have to collect all rows of this psm, and merge + write them later!
#                 rows_to_merge_dict[psm].append(row)
#             else:
#                 raise Exception("This should never happen.")
#         # finished parsing the old unmerged unified csv
#         for rows_to_merge in rows_to_merge_dict.values():
#             writer.writerow(
#                 merge_rowdicts(rows_to_merge, joinchar=joinchar)
#             )
#     os.remove(tmp_file)  # remove the old unified csv that contains duplicate rows



if __name__ == '__main__':
    if len(sys.argv) < 7:
        print(__doc__)
        sys.exit(1)

    scan_rt_lookup = pickle.load(open(sys.argv[3], 'rb'))

    params = {
        'translations' : {
            # 'aa_exception_dict' : {
            #     'J' : {
            #     'original_aa' : 'L',
            #     },
            #     'O' : {
            #         'original_aa' : 'K',
            #         'unimod_name' : 'Methylpyrroline',
            #     },
            # },
            'mods' : {
                'fix': {
                    '_id': 1,
                    'aa': 'C',
                    'composition': {'C': 2, 'H': 3, 'N': 1, 'O': 1},
                    'id': '4',
                    'mass': 57.021464,
                    'name': 'Carbamidomethyl',
                    'org': 'C,fix,any,Carbamidomethyl',
                    'pos': 'any',
                    'unimod': True
                },
                'opt': {
                    '_id': 3,
                    'aa': 'M',
                    'composition': {'O': 1},
                    'id': '35',
                    'mass': 15.994915,
                    'name': 'Oxidation',
                    'org': 'M,opt,any,Oxidation',
                    'pos': 'any',
                    'unimod': True
                },
            },
            'protein_delimiter'        : '<|>',
            'enzyme'                   : 'KR;C;P',
            'keep_asp_pro_broken_peps' : True,
            'semi_enzyme'              : False,
            'decoy_tag'                : 'decoy_',
            # 'psm_merge_delimiter'      : ';',
            'precursor_mass_tolerance_plus':5,
            'precursor_mass_tolerance_minus':5,
            'precursor_isotope_range': '0,1',
            'max_missed_cleavages' : 2
        },
        # 'label'                    : '15N',
        'label' : '',
        'prefix'                   : None
    }
    params['translations']['database'] = sys.argv[6]
    # start_time = time.time()
    main(
        input_file     = sys.argv[1],
        output_file    = sys.argv[2],
        scan_rt_lookup = scan_rt_lookup,
        params         = params,
        search_engine  = sys.argv[4],
        score_colname  = sys.argv[5]
    )    # end_time = time.time()
    # print(end_time-start_time)
    # input()



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os
import sys


def main(folder):
    '''
    Executes a search with different versions of msgf+ m on an example file from the
    data from Bruderer et al.

    usage:
        ./msgf_plus_version_comparison_qexactive.py <folder containing B_D140314_SGSDSsample1_R01_MSG_T0.mzML.gz>


    Creates a Venn diagram with the peptides obtained by the different versions.


    '''

    required_example_file = 'B_D140314_SGSDSsample1_R01_MSG_T0.mzML.gz'
    full_path = os.path.join(folder, required_example_file)

    if os.path.exists(full_path) is False:

        print(
            '''
            Your specified folder does not contain the required example file:
            {0}
            The RAW data from peptideatlas.org (PASS00589, password: WF6554orn)
            will be downloaded.
            Please convert to mzML after the download has finished and run this
            script again.
            '''.format(required_example_file)
        )

        ftp_get_params = {
            'ftp_url': 'ftp.peptideatlas.org',
            'ftp_login': 'PASS00589',
            'ftp_password': 'WF6554orn',
            'ftp_include_ext': [
                required_example_file.replace(
                    '.mzML',
                    '.raw'
                )
            ],
            'ftp_output_folder': folder,
        }
        uc = ursgal.UController(
            params=ftp_get_params
        )
        uc.fetch_file(
            engine='get_ftp_files_1_0_0'
        )
        sys.exit(1)

    engine_list = [
        'msgfplus_v9979',
        'msgfplus_v2016_09_16',
        'msgfplus_v2018_01_30',
        'msgfplus_v2018_06_28',
        'msgfplus_v2018_09_12',
    ]

    params = {
        'database': os.path.join(
            os.pardir,
            'example_data',
            'hs_201303_qs_sip_target_decoy.fasta'
        ),
        'modifications': ['C,fix,any,Carbamidomethyl'],
        'csv_filter_rules': [
            ['PEP', 'lte', 0.01],
            ['Is decoy', 'equals', 'false']
        ],
        'http_url': 'http://www.uni-muenster.de/Biologie.IBBP.AGFufezan/misc/hs_201303_qs_sip_target_decoy.fasta',
        'http_output_folder': os.path.join(
            os.pardir,
            'example_data'
        ),
        'machine_offset_in_ppm': -5e-6,
        'remove_temporary_files': False
    }

    uc = ursgal.UController(
        profile='QExactive+',
        params=params
    )

    if os.path.exists(params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )

    mzML_file = os.path.join(folder, required_example_file)

    filtered_files_list = []
    for engine in engine_list:

        unified_result_file = uc.search(
            input_file=mzML_file,
            engine=engine,
            force=False,
        )

        validated_file = uc.validate(
            input_file=unified_result_file,
            engine='percolator_2_08',
        )

        filtered_file = uc.execute_misc_engine(
            input_file=validated_file,
            engine='filter_csv_1_0_0',
        )

        filtered_files_list.append(filtered_file)

    uc.visualize(
        input_files=filtered_files_list,
        engine='venndiagram_1_1_0',
    )
    return

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(main.__doc__)
        sys.exit(1)
    main(sys.argv[1])



#!/usr/bin/env python
import ursgal
import os


class merge_csvs_1_0_0( ursgal.UNode ):
    """Merge CSVS 1_0_0 UNode"""
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Merge CSVs',
        'version'            : '1.0.0',
        'release_date'       : '2016-3-4',
        'engine_type' : {
            'misc_engine' : True,
        },
        'input_extensions'   : ['.csv'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : 'merged',
        'include_in_git'     : True,
        'in_development'     : False,
        'distributable'      : True,
        'utranslation_style' : 'merge_csvs_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'merge_csvs_1_0_0.py',
                },
            },
        },
        'citation' : \
            'Kremer, L. P. M., Leufken, J., Oyunchimeg, P., Schulze, S. & '\
            'Fufezan, C. (2016) Ursgal, Universal Python Module Combining '\
            'Common Bottom-Up Proteomics Tools for Large-Scale Analysis. J. '\
            'Proteome res. 15, 788-794.',
    }

    def __init__(self, *args, **kwargs):
        super(merge_csvs_1_0_0, self).__init__(*args, **kwargs)

    def _execute( self ):
        '''
        Merges .csv files

        for same header, new rows are appended

        for different header, new columns are appended
        '''
        print('[ -ENGINE- ] Merging csv files...')
        # self.time_point(tag = 'execution')
        csv_files = []

        for input_file_dict in self.params['input_file_dicts']:
            csv_files.append(
                os.path.join(
                    input_file_dict['dir'],
                    input_file_dict['file']
                )
            )

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        merge_csv_main = self.import_engine_as_python_function()
        merged_csv_output_path = merge_csv_main(
            csv_files = csv_files,
            output    = self.params['translations']['output_file_incl_path'],
        )
        # self.print_execution_time(tag='execution')
        return merged_csv_output_path



'''
usage: python _test_node.py -i x -o test.txt -t1 a -t2 3
'''
import argparse

'''

NOTE:
    Currently the wrapper does not call this function.

'''

def _test(**kwargs):
    '''
    '''
    with open(kwargs['output_file'], 'w') as outf:
        for key, val in kwargs.items():
            outf.write('{0},{1}\n'.format(key, val))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-i', '--input_file', type=str, required=True,
        help='Input file name')
    parser.add_argument(
        '-o', '--output_file', type=str, required=True,
        help='Output file name')
    parser.add_argument(
        '-t1', '--test1', type=str, required=True,
        help='Test parameter 1')
    parser.add_argument(
        '-t2', '--test2', type=int, required=True,
        help='Test parameter 2')

    _test(
        **vars(parser.parse_args())
    )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import unittest
import pprint
# from ursgal import umapmaster as umama
import pprint
import os
try:
    import regex
    regex_module_installed = True
except:
    regex_module_installed = False
    


TEST_FASTA = [
    '>Protein1\n',
    'ELVISLIVES\n',
    '\n',
    '>Protein2\n',
    'KLEINERPENNER\n',
    '>Protein3\n',
    'WHYISELVIS\n',
    'HELEAVING\n',
    '>T1',
    'AAAAAAA___AAAAAAAAAA_____AAAAA__________AAAAAAAAAA',
    #----+----|----+----|----+----|----+----|----+----|
    '>Overlapping',
    'GGGGGGGGGG',
    '>ugly_fasta',
    'FCKTHIS*',
    '>SortingPepCorrect_1',
    'FORWARDKKK',
    '>SortingPepCorrect_2',
    'DRAWROFKKK',
    '>SortingPepCorrect_3',
    'FROWARDKKK',
]

TEST_FASTA_TWO = [
    '>Ze_other_protein2\n',
    'KLEINERRETTER\n',
]


class UMapMaster(unittest.TestCase):
    def setUp(self):
        self.uc = ursgal.UController( verbose = False )
        upapa_class = self.uc.unodes['upeptide_mapper_1_0_0']['class'].import_engine_as_python_function('UPeptideMapper_v2')
        self.upapa_class = upapa_class( word_len=5 )
        self.upapa_class.build_lookup(
            fasta_name = 'Test.fasta',
            fasta_stream = TEST_FASTA
        )

    def test_incremental_fcache_buildup_and_global_access(self):
        '''
        This is not longer relevant, we have no global access on the peptide
        mapper any more
        '''
        # map with empty
        self.upapa_class.purge_fasta_info('Test.fasta')
        maps = self.upapa_class.map_peptide( peptide='KLEINER', fasta_name='Test.fasta')
        # print(maps)
        self.assertEqual( len(maps), 0)

        # parse one fasta via a unode ..
        self.upapa_class.build_lookup(
            fasta_name = 'Test.fasta',
            fasta_stream = TEST_FASTA
        )
        # map with one parsed fasta
        maps = self.upapa_class.map_peptide( peptide='KLEINER', fasta_name='Test.fasta')
        self.assertEqual( len(maps), 1)
        # parse another fasta via a different unode ..
        self.upapa_class.build_lookup(
            fasta_name = 'Test.fasta',
            fasta_stream = TEST_FASTA_TWO
        )
        # map with two parsed fastas

        maps = self.upapa_class.map_peptide( peptide='KLEINER', fasta_name='Test.fasta')
        print(maps)
        self.assertEqual( len(maps), 2)

    def test_fasta_id_parsed_and_available(self):
        input_fastas = []
        for line in TEST_FASTA:
            if line.startswith('>'):
                input_fastas.append( line[1:].strip() )

        self.assertEqual(
            sorted(self.upapa_class.fasta_sequences['Test.fasta'].keys()),
            sorted( input_fastas )
        )

    def test_peptide_eq_word_len(self):
        # print( self.upapa_class['Test.fasta'].keys())
        # print( ''.join(sorted('ELVIS')))
        # print( self.upapa_class['Test.fasta'][''.join(sorted('ELVIS'))] )
        self.assertEqual(
            self.upapa_class['Test.fasta'][''.join(sorted('ELVIS'))],
            set([
                ('Protein1', 1),
                ('Protein1', 5), # new srted algorithm to reduce mem print :)
                ('Protein1', 6),
                ('Protein3', 4),
                ('Protein3', 5),
                ('Protein3', 6)
            ])
        )

    def test_peptide_eq_word_len_2(self):
        maps = self.upapa_class.map_peptide( peptide='VISHE', fasta_name='Test.fasta')
        self.assertEqual( len(maps), 1 )
        for mapping in maps:
                self.assertEqual(
                    'Protein3',
                    mapping['id']
                )

    def test_peptide_lt_word_len(self):
        expected = {
            'Protein1' : {
                'pre'   : 'V',
                'post'  : 'L',
                'start' : 4,
                'end'   : 5,
                'id'    : 'Protein1'
            },
            'ugly_fasta': {
                'pre'   : 'H',
                'post'  : '-',
                'start' : 6,
                'end'   : 7,
                'id'    : 'ugly_fasta'
            },

        }
        maps = self.upapa_class.map_peptide( peptide='IS', fasta_name='Test.fasta')
        self.assertEqual( len(maps), 4 )
        for mapping in maps:
            if mapping['id'] == 'Protein1':
                self.assertEqual(
                    expected['Protein1'],
                    mapping
                )
            if mapping['id'] == 'ugly_fasta':
                self.assertEqual(
                    expected['ugly_fasta'],
                    mapping
                )

    def test_peptide_gt_word_len(self):
        expected = {
            'Protein1' : {
                'pre'   : 'L',
                'post'  : '-',
                'start' : 3,
                'end'   : 10,
                'id'    : 'Protein1'
            }
        }
        maps = self.upapa_class.map_peptide( peptide='VISLIVES', fasta_name='Test.fasta')
        self.assertEqual( len(maps), 1 )
        for mapping in maps:
            if mapping['id'] == 'Protein1':
                self.assertEqual(
                    expected['Protein1'],
                    mapping
                )

    def test_purge_fasta(self):
        self.upapa_class.purge_fasta_info( 'Test.fasta')
        self.assertFalse(
            'Test.fasta' in self.upapa_class.keys()
        )

    def test_peptide_gt_word_but_not_continous(self):
        self.assertEqual(
            self.upapa_class.map_peptide(
                peptide='WHYELVIS',
                fasta_name='Test.fasta'
            ),
            []
        )

    def test_multiple_occurrence_in_one_seq(self):
        maps = self.upapa_class.map_peptide( peptide='AAAAAAAAAA', fasta_name='Test.fasta')
        self.assertEqual( len(maps), 2 )
        sorted_maps = sorted( maps, key= lambda x: x['start'])
        self.assertEqual(
            sorted( maps, key= lambda x: x['start']),
            sorted(
                [
                    {
                        'pre'   : '_',
                        'post'  : '_',
                        'start' : 11,
                        'end'   : 20,
                        'id'    : 'T1'
                    },
                    {
                        'pre'   : '_',
                        'post'  : '-',
                        'start' : 41,
                        'end'   : 50,
                        'id'    : 'T1'
                    }

                ], key= lambda x: x['start']
            )
        )

    def test_multiple_occurrence_with_opverlap_in_one_seq(self):
        maps = self.upapa_class.map_peptide(
            peptide='GGGGGGG',
            fasta_name='Test.fasta'
        )
        self.assertEqual( len(maps), 4 )
        self.assertEqual(
            sorted([ m['start'] for m in maps] ),
            [1,2,3,4]
        )
        if regex_module_installed:
            #test short sequence, regex does not work with overlap
            maps = self.upapa_class.map_peptide(
                peptide='GGGG',
                fasta_name='Test.fasta'
            )
            self.assertEqual( len(maps), 7 )

    def test_sort_independece(self):
        map_1 = self.upapa_class.map_peptide(
            peptide    = 'FORWARD',
            fasta_name = 'Test.fasta'
        )
        map_2 = self.upapa_class.map_peptide(
            peptide    = 'DRAWROF',
            fasta_name = 'Test.fasta'
        )
        map_3 = self.upapa_class.map_peptide(
            peptide    = 'FORWA',
            fasta_name = 'Test.fasta'
        )
        map_4 = self.upapa_class.map_peptide(
            peptide    = 'AWROF',
            fasta_name = 'Test.fasta'
        )
        map_5 = self.upapa_class.map_peptide(
            peptide    = 'AORRW',
            fasta_name = 'Test.fasta'
        ) # should not work..

        self.assertEqual(len(map_1), 1)
        self.assertEqual(len(map_2), 1)
        self.assertEqual(len(map_3), 1)
        self.assertEqual(len(map_4), 1)
        self.assertEqual(len(map_5), 0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
'''
Merges csvs into one single one.

Usage:
./merge_csvs_1_0_0.py <mzML File1> ... <mzML FileN>

Note: All headers have to be the same, otherwise warning will be printed
'''

from __future__ import print_function
import sys
import os
import csv

# csv.field_size_limit(sys.maxsize)


def main(csv_files=None, output=None):
    '''Merges ident csvs'''
    assert output is not None, 'merge_csv requires output file name'
    buffered_input_files = {}
    all_fieldnames = []
    all_fieldnames_the_same = True
    for i, csv_file in enumerate( csv_files ):
        file_object = open( csv_file , 'r' )
        buffered_input_files[ csv_file ] = csv.DictReader(
            row for row in file_object if not row.startswith('#')
        )
        for fieldname in buffered_input_files[ csv_file ].fieldnames:
            if fieldname not in all_fieldnames:
                all_fieldnames.append( fieldname )
                if i != 0:
                    all_fieldnames_the_same = False

    if all_fieldnames_the_same is False:
        print('''
    WARNING!
    WARNING!   Not all fieldnames / headers of the input csvs were the same!
    WARNING!
        ''')
    out_csv = open( output, 'w')
    csv_kwargs = {}
    if sys.platform == 'win32':
        csv_kwargs['lineterminator'] = '\n'
    else:
        csv_kwargs['lineterminator'] = '\r\n'
    csv_writer = csv.DictWriter(
        out_csv,
        all_fieldnames,
        **csv_kwargs
    )
    csv_writer.writeheader()

    for csv_file, csv_dict_reader in buffered_input_files.items():
        for line_dict in csv_dict_reader:
            csv_writer.writerow( line_dict )
    out_csv.close()
    return None


if __name__ == '__main__':
    # parsing command line arguments:
    if len(sys.args) == 1:
        print(__doc__)
        sys.exit(1)
    else:
        main(
            csv_files = sys.argv[1:],
            output = 'merged_csv.py'
        )



#!/usr/bin/env python
# encoding: utf-8
'''

Test the csv2counted_results engine for unified file

'''
import ursgal
import csv
import os


R = ursgal.UController()

csv2counted_results_main = R.unodes['csv2counted_results_1_0_0']['class'].import_engine_as_python_function()
input_csv = os.path.join(
    'tests',
    'data',
    'csv2counted_results_1_0_0',
    'csv2counted_results_test.csv'
)
output_csv = os.path.join(
    'tests',
    'data',
    'csv2counted_results_1_0_0',
    'csv2counted_results_test_counted.csv'
)
csv2counted_results_main(
    input_file     = input_csv,
    output_file    = output_csv,
    identifier_colum_names = ['Protein ID', 'Charge'],
    count_column_names = ['Sequence', 'Modifications'],
    count_by_file = True,
)

correct_values_dict = {
    'ProteinA#3' : {
        'test' : '3',
        'test2' : '0',
    },
    'ProteinA#2' : {
        'test' : '1',
        'test2' : '0',
    },
    'ProteinC#3' : {
        'test' : '1',
        'test2' : '0',
    },
    'ProteinC#2' : {
        'test' : '2',
        'test2' : '0',
    },
    'ProteinD#2' : {
        'test' : '0',
        'test2' : '1',
    },
    'ProteinD#1' : {
        'test' : '0',
        'test2' : '1',
    },
}

test_dict = {}
csv_out = csv.DictReader(open(output_csv, 'r'))
for line_dict in csv_out:
    test_dict[
        '#'.join([
            line_dict['Protein ID'],
            line_dict['Charge']
        ])
    ] = {
        'test' : line_dict['test'],
        'test2' : line_dict['test2'],
    }

def csv2counted_results_test():
    for test_id, test_key in enumerate(correct_values_dict.keys()):
        yield csv2counted_results, test_key, test_dict


def csv2counted_results( test_key, test_dict ):
    assert os.path.isfile(output_csv)

    print(test_key, correct_values_dict[test_key], test_dict[test_key])
    assert correct_values_dict[test_key] == test_dict[test_key]


if __name__ == '__main__':
    print(__doc__)
    for test_id, test_key in enumerate(correct_values_dict.keys()):
        csv2counted_results(test_key, test_dict)



#!/usr/bin/env python
import ursgal
import os


class get_http_files_1_0_0( ursgal.UNode ):
    """
    get_http_files_1_0_0 UNode

    Downloads files via http

    Args:
        * http_url
        * http_output_folder
        
    Note:
        meta info param 'output_extensions' is by default txt, so that the 
        temporary txt json files get properly deleted

    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Get HTTP Files',
        'version'            : '1.0.0',
        'release_date'       : '2016-3-4',
        'engine_type' : {
            'fetcher' : True,
        },
        'input_extensions'   : [],
        'output_extensions'  : ['.txt'],
        'output_suffix'      : None,
        'in_development'     : False,
        'include_in_git'     : True,
        'distributable'      : True,
        'utranslation_style' : 'get_http_style_1',
        # 'engine_url' : {
        #     'internal' : True,
        # },
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'get_http_files_1_0_0.py',
                },
            },
        },
        'citation' : \
            'Kremer, L. P. M., Leufken, J., Oyunchimeg, P., Schulze, S. & '\
            'Fufezan, C. (2016) Ursgal, Universal Python Module Combining '\
            'Common Bottom-Up Proteomics Tools for Large-Scale Analysis. J. '\
            'Proteome res. 15, 788-794.',
    }

    def __init__(self, *args, **kwargs):
        super(get_http_files_1_0_0, self).__init__(*args, **kwargs)

    def _execute( self ):
        '''
        Downloads files via http

        '''
        print('[ -ENGINE- ] Executing HTTP Download ..')
        # self.time_point(tag = 'execution')
        main = self.import_engine_as_python_function()
        output_path = main(
            http_url            = self.params.get('http_url', None),
            http_output_folder  = self.params.get('http_output_folder', None),
        )
        # self.print_execution_time(tag='execution')
        self.io['output']['finfo']['dir']  = os.path.dirname( output_path )
        self.io['output']['finfo']['file'] = os.path.basename( output_path )
        return output_path



#!/usr/bin/env python
import ursgal


class mascot_x_x_x( ursgal.UNode ):
    """
    Dummy to merge mascot data into usgal workflow
    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'Mascot',
        'version'                     : '2017',
        'engine_type' : {
            'protein_database_search_engine' : True,
        },
        'output_extensions'           : ['.xml'],
        'create_own_folder'           : True,
        'utranslation_style'          : 'mascot_style_1',
        'citation'                    : 'www.matrixscience.com',
        'in_development'              : True,
        'include_in_git'              : True,
        'input_extensions'            : [],
        'release_date'                : None
    }

    def __init__(self, *args, **kwargs):
        super(mascot_x_x_x, self).__init__(*args, **kwargs)



#!/usr/bin/env python
import ursgal
import os

from .xtandem_vengeance import xtandem_vengeance as xtandem

class xtandem_alanine( xtandem ):
    """
    X!Tandem UNode
    Parameter options at http://www.thegpm.org/TANDEM/api/

    Reference:
    Craig R, Beavis RC. (2004) TANDEM: matching proteins with tandem mass spectra.
    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'X!Tandem',
        'version'                     : 'ALANINE',
        'release_date'                : '2017-02-01',
        'engine_type' : {
            'protein_database_search_engine' : True,
        },
        'input_extensions'            : ['.mgf', '.gaml', '.dta', '.pkl', '.mzData', '.mzXML'],
        'output_extensions'           : ['.xml'],
        'create_own_folder'           : True,
        'in_development'              : False,
        'include_in_git'              : False,
        'distributable'               : False,
        'utranslation_style'          : 'xtandem_style_1',
        'engine' : {
            'darwin' : {
                '64bit' : {
                    'exe'            : 'tandem',
                    'url'            : '',
                    'zip_md5'        : 'f33da624f5bdd2702c60e4dc86e9137c',
                    'additional_exe' : [],
                },
            },
            'linux' : {
                '64bit' : {
                    'exe'            : 'tandem.exe',
                    'url'            : '',
                    'zip_md5'        : '37f087c213dfed3c95e62a865d843fb9',
                    'additional_exe' : [],
                },
            },
            'win32' : {
                '64bit' : {
                    'exe'            : 'tandem.exe',
                    'url'            : '',
                    'zip_md5'        : '78ed3c76271f3596aceea8ea5bc904bc',
                    'additional_exe' : [],
                },
            },
        },
        'citation' : \
            'Craig R, Beavis RC. (2004) TANDEM: matching proteins with tandem '\
            'mass spectra.',
    }


#!/usr/bin/env python3
# encoding: utf-8
'''

Test the OMSSA mop remap of e.g. TMT labeling

TMT labeling has the wrong unimod name in the OMSSA mods xml file

Probably more mods need remap

'K,fix,any,TMT6plex',
'*,opt,N-term,TMT6plex',
'K,fix,any,TMT2plex',
'*,opt,N-term,TMT2plex',

'''
import ursgal
import csv
import os
import pprint


TESTS = [
    {
        'modifications' : [ 'K,fix,any,TMT6plex' ],
        'result_dict': {
            '737' : {
                '198' : {
                    'aa_targets'  : ['K'],
                }
            }
        },
    },
    {
        'modifications' : [ '*,opt,N-term,TMT6plex' ],
        'result_dict': {
            '737' : {
                '199' : {
                    'aa_targets'  : ['N-term'],
                }
            }
        },
    },
    {
        'modifications' : [ 'K,fix,any,TMT2plex' ],
        'result_dict': {
            '738' : {
                '198' : {
                    'aa_targets'  : ['K'],
                }
            }
        },
    },
    {
        'modifications' : [ '*,opt,N-term,TMT2plex' ],
        'result_dict': {
            '738' : {
                '199' : {
                    'aa_targets'  : ['N-term'],
                }
            }
        },
    },

]

uc = ursgal.UController()
IS_AVAILABLE = uc.unodes['omssa_2_1_9']['available']

def map_mods_test():
    for test_id, test_dict in enumerate(TESTS):
        yield map_mods, test_dict


def map_mods( test_dict ):
    if IS_AVAILABLE:
        uc.params['modifications'] = test_dict['modifications']
        omssa_class = cl=uc.unodes['omssa_2_1_9']['_wrapper_class']()
        cl.exe = os.path.join(uc.unodes['omssa_2_1_9']['resource_folder'],'omssacl')
        cl._load_omssa_xml()
        for unimod_id in test_dict['result_dict'].keys():
            assert unimod_id in cl.omssa_mod_Mapper.keys()
            for omssa_id in test_dict['result_dict'][unimod_id].keys():
                assert omssa_id in cl.omssa_mod_Mapper[unimod_id].keys()
                for k,v in test_dict['result_dict'][unimod_id][omssa_id].items():
                    assert cl.omssa_mod_Mapper[unimod_id][omssa_id][k] == v


if __name__ == '__main__':
    print(__doc__)
    for test_id, test_dict in enumerate(TESTS):
        map_mods(test_dict)



#!/usr/bin/env python
import ursgal
import os
import sys

class msgfplus_v9979( ursgal.UNode ):
    """
    MSGF+ UNode
    Parameter options at https://bix-lab.ucsd.edu/pages/viewpage.action?pageId=13533355

    Reference:
    Kim S, Mischerikow N, Bandeira N, Navarro JD, Wich L, Mohammed S, Heck AJ, Pevzner PA. (2010) The Generating Function of CID, ETD, and CID/ETD Pairs of Tandem Mass Spectra: Applications to Database Search.
    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'MSGF+',
        'version'                     : 'v9979',
        'release_date'                : '2010-12-1',
        'engine_type' : {
            'protein_database_search_engine' : True,
        },
        'input_extensions'            : ['.mgf', '.mzML', '.mzXML', '.ms2', '.pkl', '.dta.txt'],
        'output_extensions'           : ['.mzid'],
        'create_own_folder'           : True,
        'in_development'              : False,
        'include_in_git'              : False,
        'distributable'                : True,
        'utranslation_style'          : 'msgfplus_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'            : 'MSGFPlus.jar',
                    'url'            : 'http://proteomics.ucsd.edu/Software/MSGFPlus/MSGFPlus.zip',
                    'zip_md5'        : '5eeb2f74f708f1861ca570a46e8a051d',
                    'additional_exe' : [],
                },
            },
        },
        'citation' : \
            'Kim S, Mischerikow N, Bandeira N, Navarro JD, Wich L, '\
            'Mohammed S, Heck AJ, Pevzner PA. (2010) The Generating Function '\
            'of CID, ETD, and CID/ETD Pairs of Tandem Mass Spectra: '\
            'Applications to Database Search.',
    }

    def __init__(self, *args, **kwargs):
        super(msgfplus_v9979, self).__init__(*args, **kwargs)
        pass

    def preflight( self ):
        '''
        Formatting the command line via self.params

        Modifications file will be created in the output folder

        Returns:
                dict: self.params
        '''

        translations = self.params['translations']['_grouped_by_translated_key']
        # import pprint
        # pprint.pprint(translations)
        # exit(1)

        self.params[ 'command_list' ] = [
            'java',
            '-jar',
            self.exe,
        ]

        self.params['translations']['mgf_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        translations['-s']['mgf_input_file'] = self.params['translations']['mgf_input_file']

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        translations['-o']['output_file_incl_path'] = self.params['translations']['output_file_incl_path']

        self.params['translations']['modification_file'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file'] + '_Mods.txt'
        )
        self.created_tmp_files.append( self.params['translations']['modification_file'] )
        translations['-mod']['modifications'] = self.params['translations']['modification_file']

        mods_file = open( self.params['translations']['modification_file'], 'w', encoding = 'UTF-8' )
        modifications = []

        print('NumMods={0}'.format(translations['NumMods']['max_num_mods']), file = mods_file)

        if self.params['translations']['label'] == '15N':
            for aminoacid, N15_Diff in ursgal.ukb.DICT_15N_DIFF.items():
                existing = False
                for mod in self.params[ 'mods' ][ 'fix' ]:
                    if aminoacid == mod[ 'aa' ]:
                        mod[ 'mass' ] += N15_Diff
                        mod[ 'name' ] += '_15N_{0}'.format(aminoacid)
                        existing = True
                if existing == True:
                    continue
                else:
                    modifications.append( '{0},{1},fix,any,15N_{1}'.format( N15_Diff, aminoacid ) )

        for t in [ 'fix', 'opt' ]:
            for mod in self.params[ 'mods' ][ t ]:
                modifications.append( '{0},{1},{2},{3},{4}'.format(mod[ 'mass' ], mod[ 'aa' ], t, mod[ 'pos' ], mod[ 'name' ] ) )

        for mod in modifications:
            print( mod, file = mods_file )

        mods_file.close()

        translations['-t'] = {
            '-t' : '{0}{1}, {2}{1}'.format(
                translations['-t']['precursor_mass_tolerance_minus'],
                translations['-t']['precursor_mass_tolerance_unit'],
                translations['-t']['precursor_mass_tolerance_plus'],
            )
        }

        command_dict = {}

        for translated_key, translation_dict in translations.items():
            if translated_key == '-Xmx':
                self.params[ 'command_list' ].insert(1,'{0}{1}'.format(
                    translated_key,
                    list(translation_dict.values())[0]
                ))
            elif translated_key in ['label', 'NumMods']:
                continue
            elif len(translation_dict) == 1:
                command_dict[translated_key] = str(list(translation_dict.values())[0])
            else:
                print('The translatd key ', translated_key, ' maps on more than one ukey, but no special rules have been defined')
                print(translation_dict)
                sys.exit(1)
        for k, v in command_dict.items():
            self.params[ 'command_list' ].extend((k, v))

        return self.params



#!/usr/bin/env python
'''
Convert unified .csv files to .ssl files.

Resulting .ssl files can be used as input for BiblioSpec

usage:
    ./csv2ssl_1_0_0.py <input_file> <output_file> <score_column_name> <score_type>

'''

import sys
import os
import csv
import ursgal
import re

if sys.platform != 'win32':
    csv.field_size_limit(sys.maxsize)


def main(
    input_file=None,
    output_file=None,
    score_column_name=None,
    score_type=None
):
    '''
    Convert csvs to ssl
    '''

    umama = ursgal.UnimodMapper()

    csv_kwargs = {}
    if sys.platform == 'win32':
        csv_kwargs['lineterminator'] = '\n'
    else:
        csv_kwargs['lineterminator'] = '\r\n'

    output_file_object = open(output_file,'w')
    new_fieldnames = [
        'file',
        'scan',
        'charge',
        'sequence',
        'score-type',
        'score',
        'retention-time'
    ]

    with open( input_file, 'r' ) as in_file:
        csv_input  = csv.DictReader( in_file )
        csv_output = csv.DictWriter(
            output_file_object,
            new_fieldnames,
            delimiter='\t',
            **csv_kwargs
        )
        csv_output.writeheader()

        for csv_line_dict in csv_input:
            rt = round(float(csv_line_dict['Retention Time (s)'])/60, 10)
            sequence = csv_line_dict['Sequence']
            mods = csv_line_dict['Modifications'].split(';')

            pattern = re.compile( r''':(?P<pos>[0-9]*$)''' )
            pos2mass = {}
            for mod in mods:
                if mod == '':
                    continue
                if ':' not in mod:
                    sys.exit('This unimod: {0} requires positional information'.format( mod ))
                for occ, match in enumerate( pattern.finditer( mod )):
                    try:
                        unimod_mass = umama.name2mass(
                            mod[ :match.start() ]
                        )
                    except:
                        sys.exit(
                            'Can not map unimod {0}. extracted position argument: {1}'.format(
                                mod, match.start()
                        ))
                    position = int(match.group('pos'))
                    if position == 0:
                        position = 1
                    elif position > len(sequence):
                        position = len(sequence)
                    if position not in pos2mass:
                        pos2mass[position] = 0
                    pos2mass[ position ] += unimod_mass
                    if occ >= 1:
                        sys.exit(
                        'Incorrect regex pattern for mod: {0}'.format(
                            mod
                         ))

            seq_incl_mods = ''
            for p, aa in enumerate(sequence):
                seq_incl_mods += aa
                if p+1 in pos2mass:
                    mass = pos2mass[p+1]
                    if mass >= 0:
                        seq_incl_mods += '[+{0}]'.format(mass)
                    else:
                        seq_incl_mods += '[{0}]'.format(mass)

            ssl_line_dict = {}
            ssl_line_dict['file'] = csv_line_dict['Raw data location']
            ssl_line_dict['scan'] = csv_line_dict['Spectrum ID']
            ssl_line_dict['charge'] = csv_line_dict['Charge']
            ssl_line_dict['sequence'] = seq_incl_mods
            ssl_line_dict['score-type'] = score_type
            ssl_line_dict['score'] = csv_line_dict[score_column_name]
            ssl_line_dict['retention-time'] = rt

            csv_output.writerow( ssl_line_dict )

    output_file_object.close()
    in_file.close()
    return output_file

if __name__ == '__main__':
    if len(sys.argv) < 4:
        print(__doc__)
        sys.exit(1)

    if sys.argv[3] == 'None':
        output_file_unfiltered = None
    else:
        output_file_unfiltered = sys.argv[3]


    main(
        input_file     = sys.argv[1],
        output_file    = sys.argv[2],
        score_column_name = sys.argv[3],
        score_type    = sys.argv[4],
    )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os


def main():
    '''
    Executes a search with 3 different search engines on an example file from
    the data from Barth et al. Two searches are performed pe engine for each
    14N (unlabeled) and 15N labeling. The overlap of identified peptides
    for the 14N and 15N searches between the engines is visualized as well as 
    the overlap between all 14N and 15N identified peptides.

    usage:
        ./search_with_label_15N.py

    Note:
        It is important to convert the mgf file outside of (and before :) ) the 
        search loop to avoid mgf file redundancy and to assure correct retention
        time mapping in the unify_csv node.

    '''

    engine_list = [
        'omssa_2_1_9',
        'xtandem_piledriver',
        'msgfplus_v9979',
    ]

    params = {
        'database': os.path.join(
            os.pardir,
            'example_data',
            'Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta'
        ),
        'modifications': [],
        'csv_filter_rules': [
            ['PEP', 'lte', 0.01],
            ['Is decoy', 'equals', 'false']
        ],
        'ftp_url': 'ftp.peptideatlas.org',
        'ftp_login': 'PASS00269',
        'ftp_password': 'FI4645a',
        'ftp_include_ext': [
            'JB_FASP_pH8_2-3_28122012.mzML',
        ],
        'ftp_output_folder': os.path.join(
            os.pardir,
            'example_data',
            'search_with_label_15N'
        ),
        'http_url': 'https://www.sas.upenn.edu/~sschulze/Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta',
        'http_output_folder': os.path.join(
            os.pardir,
            'example_data'
        )
    }

    if os.path.exists(params['ftp_output_folder']) is False:
        os.mkdir(params['ftp_output_folder'])

    uc = ursgal.UController(
        profile='LTQ XL low res',
        params=params
    )
    mzML_file = os.path.join(
        params['ftp_output_folder'],
        params['ftp_include_ext'][0]
    )
    if os.path.exists(mzML_file) is False:
        uc.fetch_file(
            engine='get_ftp_files_1_0_0'
        )
    if os.path.exists(params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )

    mgf_file = uc.convert(
        input_file=mzML_file,
        engine='mzml2mgf_1_0_0',
    )
    files_2_merge = {}
    label_list = ['14N', '15N']
    for label in label_list:
        validated_and_filtered_files_list = []
        uc.params['label'] = label
        uc.params['prefix'] = label
        for engine in engine_list:
            search_result = uc.search_mgf(
                input_file=mgf_file,
                engine=engine,
            )
            converted_result = uc.convert(
                input_file=search_result,
                guess_engine=True,
            )
            mapped_results = uc.execute_misc_engine(
                input_file=converted_result,
                engine='upeptide_mapper',
            )
            unified_search_results = uc.execute_misc_engine(
                input_file=mapped_results,
                engine='unify_csv'
            )
            validated_file = uc.validate(
                input_file=unified_search_results,
                engine='percolator_2_08',
            )
            filtered_file = uc.execute_misc_engine(
                input_file=validated_file,
                engine='filter_csv'
            )
            validated_and_filtered_files_list.append(filtered_file)
        files_2_merge[label] = validated_and_filtered_files_list
        uc.visualize(
            input_files=validated_and_filtered_files_list,
            engine='venndiagram_1_1_0',
        )
    uc.params['prefix'] = None
    uc.params['label'] = ''
    uc.params['visualization_label_positions'] = {}
    label_comparison_file_list = []
    for n, label in enumerate(label_list):
        uc.params['visualization_label_positions'][str(n)] = label
        label_comparison_file_list.append(
            uc.execute_misc_engine(
                input_file=files_2_merge[label],
                engine='merge_csvs'
            )
        )
    uc.visualize(
        input_files=label_comparison_file_list,
        engine='venndiagram_1_1_0',
    )

    return

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import sys
import os
import pyGCluster


def main():
    '''
    Please visit pyGCluster documentation for more information on this plotting
    function:
        http://pygcluster.github.io/usage.html#clustered-data-visualization

        - color gradients
        - box styles

    Please do not forget to cite pyGCluster and Ursgal when using this node.

    Usage:

        ./simple_heatmap_example.py <path_to_csv_file>

    Note:

        Use of force = True is recommended to cover changes in the csv input
        file.
        Default value suffix of the column name is '_mean' and '_std' for the
        error estimate.
        Please refer to the documentation for further details on parameters.

    '''
    uc = ursgal.UController(
        profile='LTQ XL low res',
        params={
            'heatmap_column_positions' : {},
            'heatmap_annotation_field_name': 'map to uniprot',
            'heatmap_identifier_field_name': 'Protein',
            'heatmap_max_value': 3,
            'heatmap_min_value': -3,
            'heatmap_color_gradient': 'RdBu',
            'heatmap_box_style': 'classic'
        }
    )
    uc.visualize(
        input_files=sys.argv[1],
        engine='plot_pygcluster',
        output_file_name='{0}_heatmap.svg'.format(
            os.path.basename(sys.argv[1])
        ),
        multi=False,
        force=True
    )
    return


if __name__ == '__main__':
    if len(sys.argv) == 1:
        print(main.__doc__)
        sys.exit(1)
    main()



#!/usr/bin/env python

import ursgal
import os.path


class combine_FDR_0_1( ursgal.UNode ):
    '''
    combine FDR 0_1 UNode

    An implementation of the "combined FDR Score" algorithm, as described in:
    Jones AR, Siepen JA, Hubbard SJ, Paton NW (2009):
    "Improving sensitivity in proteome studies by analysis of false discovery
    rates for multiple search engines."

    Input should be multiple CSV files from different search engines. Each
    CSV requires a PEP column, for instance by post-processing with Percolator.

    Returns a merged CSV file with all PSMs that were found and an added
    column "Combined FDR Score".
    '''

    META_INFO = {
        'edit_version'          : 1.00,
        'name'                  : 'combine FDR',
        'version'               : '0.1',
        'release_date'          : '2009-5-1',
        'engine_type' : {
            'meta_engine' : True,
        },
        'input_extensions'      : ['.csv'],
        'output_extensions'     : ['.csv'],
        'create_own_folder'     : False,
        'in_development'        : False,
        'include_in_git'        : True,
        'distributable'      : True,
        'utranslation_style'    : 'combine_FDR_style_1',
        'engine': {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'     : 'combine_FDR_0_1.py',
                },
            },
        },
        'citation' : \
            'An implementation of the "combined FDR Score" algorithm, as '\
            'described in: Jones AR, Siepen JA, Hubbard SJ, Paton NW (2009) '\
            'Improving sensitivity in proteome studies by analysis of false '\
            'discovery rates for multiple search engines.',
    }

    def __init__(self, *args, **kwargs):
        super(combine_FDR_0_1, self).__init__(*args, **kwargs)
        pass


    def preflight( self ):
        '''
        Building the list of parameters that will be passed to the
        combine_FDR_0_1 main function.

        These parameters are stored in self.command_dict

        Returns:
                None
        '''
        input_file_list_for_cFDR_script = []
        file_info_for_cFDR_script       = {}

        input_file_search_engines = set()
        for input_file_dict in self.params["input_file_dicts"]:
            input_file_path = os.path.join(
                input_file_dict['dir'],
                input_file_dict['file']
            )
            input_file_list_for_cFDR_script.append( input_file_path )

            search_engine = input_file_dict["last_engine"]
            score_colname = input_file_dict["last_engine_colname"]
            input_file_search_engines.add( search_engine )

            file_info_for_cFDR_script[ input_file_path ] = {
                "engine_name"    : search_engine,                # i.e. "xtandem_sledgehammer"
                "engine_colname" : score_colname.split(":")[0],  # i.e. "X\!Tandem"
            }

        assert len(input_file_search_engines) == len(self.params["input_file_dicts"]) \
        and len(self.params["input_file_dicts"]) > 1, '''
        All {0} input files must be search results from *different* search engines.
        You specified {1} input files:
        {2}
        ...but they are all from {3} engine(s).
        If want to analyze multiple files from the same engine, merge them beforehand with
        merge_csvs().'''.format(
            os.path.basename( self.exe ),
            len(self.params["input_file_dicts"]),
            input_file_list_for_cFDR_script,
            len(input_file_search_engines)
        )

        self.command_dict = {
            "input_file_list" : input_file_list_for_cFDR_script,
            "cutoff"          : None,
            "directory"       : self.params['output_dir_path'],
            "file_info"       : file_info_for_cFDR_script,
            "output_filename" : self.params['output_file'],
        }


    def _execute( self ):
        '''
        Executing the combine_FDR_0_1 main function with parameters
        that were defined in preflight (stored in self.command_dict)

        The main function is imported and then executed using the
        parameters from command_dict.

        Returns:
                None
        '''
        combine_FDR_main = self.import_engine_as_python_function()

        print('''

Executing main() function from {scriptpath}.py with the following parameters:

>>> {script}.main(
...
...     input_file_list = \t{input_file_list},
...
...     file_info       = \t{file_info},
...
...     directory       = {directory},
...     output_filename = {output_filename},
... )

        '''.format(
            scriptpath = os.path.relpath( self.exe ),
            script     = os.path.basename( self.exe ),
            **self.command_dict
            )
        )

        # executing main function of the FDR combiner script
        combine_FDR_main(
            input_file_list = self.command_dict["input_file_list"],
            file_info       = self.command_dict["file_info"],
            cutoff          = 0.01,  # only for printing
            directory       = self.command_dict["directory"],
            output_filename = self.command_dict["output_filename"],
            filter_decoys   = False,  # we have filter_csv for that purpose :)
            filter_cutoff   = False,  # we have filter_csv for that purpose :)
        )
    #self.execute_answer = self.command_dict["output_filename"]
